from collections.abc import Iterator
from inspect import isclass
from typing import Optional

import _pytest.logging
import pytest
from pytest_container import Container
from pytest_container.container import ContainerData

import tmt.log
import tmt.package_managers
import tmt.package_managers.apk
import tmt.package_managers.apt
import tmt.package_managers.dnf
import tmt.package_managers.rpm_ostree
import tmt.plugins
import tmt.steps.provision.podman
import tmt.utils
from tmt.package_managers import (
    FileSystemPath,
    Installable,
    Options,
    Package,
    PackageManager,
    PackageManagerClass,
    PackagePath,
)
from tmt.steps.provision.podman import GuestContainer
from tmt.utils import ShellScript

from . import MATCH, assert_log

# We will need a logger...
logger = tmt.Logger.create()
logger.add_console_handler()

# Explore available plugins
tmt.plugins.explore(logger)

# Local images created via `make images/test`, reference to local registry
from . import (  # noqa: E402
    CONTAINER_ALPINE,
    CONTAINER_CENTOS_7,
    CONTAINER_CENTOS_STREAM_9,
    CONTAINER_CENTOS_STREAM_10,
    CONTAINER_DEBIAN_127,
    CONTAINER_FEDORA_40,
    CONTAINER_FEDORA_41,
    CONTAINER_FEDORA_COREOS,
    CONTAINER_FEDORA_COREOS_OSTREE,
    CONTAINER_FEDORA_RAWHIDE,
    CONTAINER_UBI_8,
    CONTAINER_UBUNTU_2204,
)

PACKAGE_MANAGER_DNF5 = tmt.package_managers._PACKAGE_MANAGER_PLUGIN_REGISTRY.get_plugin('dnf5')
PACKAGE_MANAGER_DNF = tmt.package_managers._PACKAGE_MANAGER_PLUGIN_REGISTRY.get_plugin('dnf')
PACKAGE_MANAGER_YUM = tmt.package_managers._PACKAGE_MANAGER_PLUGIN_REGISTRY.get_plugin('yum')
PACKAGE_MANAGER_APT = tmt.package_managers._PACKAGE_MANAGER_PLUGIN_REGISTRY.get_plugin('apt')
PACKAGE_MANAGER_RPMOSTREE = tmt.package_managers._PACKAGE_MANAGER_PLUGIN_REGISTRY.get_plugin(
    'rpm-ostree'
)
PACKAGE_MANAGER_APK = tmt.package_managers._PACKAGE_MANAGER_PLUGIN_REGISTRY.get_plugin('apk')


def has_legacy_dnf(container: ContainerData) -> bool:
    """
    Checks whether a container provides older ``dnf`` and ``yum``.

    At some point, Fedora switched to ``dnf5`` completely, and older
    ``dnf`` and ``yum`` commands are now mere symlinks to ``dnf5``.
    """

    if 'fedora' not in container.image_url_or_id and 'centos' not in container.image_url_or_id:
        return False

    return container.image_url_or_id not in (
        CONTAINER_FEDORA_RAWHIDE.url,
        CONTAINER_FEDORA_41.url,
        CONTAINER_FEDORA_COREOS.url,
        CONTAINER_FEDORA_COREOS_OSTREE.url,
    )


def has_dnf5_preinstalled(container: ContainerData) -> bool:
    """
    Checks whether a container provides ``dnf5``
    """

    return container.image_url_or_id in (
        CONTAINER_FEDORA_RAWHIDE.url,
        CONTAINER_FEDORA_41.url,
        CONTAINER_FEDORA_COREOS.url,
        CONTAINER_FEDORA_COREOS_OSTREE.url,
    )


def assert_output(
    expected_output: Optional[str], stdout: Optional[str], stderr: Optional[str]
) -> None:
    """
    Check that the expected output is present

    We don't care whether the expected string is in stdout or stderr.
    Just make sure the output is there.
    """

    # Nothing to do if there are no expectations
    if not expected_output:
        return

    combined_output = (stdout or "") + (stderr or "")
    assert combined_output != ""
    assert expected_output in combined_output


# Note: keep the list ordered by the most desired package manager to the
# least desired one. For most of the tests, the order is not important,
# but the list is used to generate the discovery tests as well, and the
# order is used to find out what package manager is expected to be
# discovered.
CONTAINER_BASE_MATRIX = [
    # Fedora
    (CONTAINER_FEDORA_RAWHIDE, PACKAGE_MANAGER_DNF5),
    (CONTAINER_FEDORA_41, PACKAGE_MANAGER_DNF5),
    (CONTAINER_FEDORA_40, PACKAGE_MANAGER_DNF5),
    (CONTAINER_FEDORA_40, PACKAGE_MANAGER_DNF),
    (CONTAINER_FEDORA_40, PACKAGE_MANAGER_YUM),
    # CentOS Stream
    (CONTAINER_CENTOS_STREAM_10, PACKAGE_MANAGER_DNF),
    (CONTAINER_CENTOS_STREAM_9, PACKAGE_MANAGER_DNF),
    (CONTAINER_CENTOS_STREAM_9, PACKAGE_MANAGER_YUM),
    # CentOS
    (CONTAINER_CENTOS_7, PACKAGE_MANAGER_YUM),
    # UBI
    (CONTAINER_UBI_8, PACKAGE_MANAGER_DNF),
    (CONTAINER_UBI_8, PACKAGE_MANAGER_YUM),
    # Ubuntu
    (CONTAINER_UBUNTU_2204, PACKAGE_MANAGER_APT),
    # Debian
    (CONTAINER_DEBIAN_127, PACKAGE_MANAGER_APT),
    # Fedora CoreOS
    (CONTAINER_FEDORA_COREOS, PACKAGE_MANAGER_DNF5),
    (CONTAINER_FEDORA_COREOS_OSTREE, PACKAGE_MANAGER_RPMOSTREE),
    # Alpine
    (CONTAINER_ALPINE, PACKAGE_MANAGER_APK),
]

CONTAINER_MATRIX_IDS = [
    f'{container.url} / {package_manager_class.__name__.lower()}'
    for container, package_manager_class in CONTAINER_BASE_MATRIX
]


CONTAINER_DISCOVERY_MATRIX: dict[str, tuple[Container, PackageManagerClass]] = {}

for container, package_manager_class in CONTAINER_BASE_MATRIX:
    if container.url in CONTAINER_DISCOVERY_MATRIX:
        continue

    CONTAINER_DISCOVERY_MATRIX[container.url] = (container, package_manager_class)


def is_dnf5_preinstalled(container: ContainerData) -> bool:
    return container.image_url_or_id in (
        CONTAINER_FEDORA_RAWHIDE.url,
        CONTAINER_FEDORA_41.url,
        CONTAINER_FEDORA_COREOS.url,
        CONTAINER_FEDORA_COREOS_OSTREE.url,
    )


def create_package_manager(
    container: ContainerData,
    guest: GuestContainer,
    package_manager_class: PackageManagerClass,
    logger: tmt.log.Logger,
) -> PackageManager:
    guest_data = tmt.steps.provision.podman.PodmanGuestData(
        image=container.image_url_or_id, container=container.container_id
    )

    guest = tmt.steps.provision.podman.GuestContainer(
        logger=logger, data=guest_data, name='dummy-container'
    )
    guest.start()
    guest.show()

    if package_manager_class is tmt.package_managers.dnf.Dnf5:
        # Note that our custom images contain `dnf5` already
        if is_dnf5_preinstalled(container):
            pass

        else:
            guest.execute(ShellScript('dnf install --nogpgcheck -y dnf5'))

    return package_manager_class(guest=guest, logger=logger)


def _parametrize_test_discovery() -> Iterator[tuple[ContainerData, PackageManagerClass]]:
    yield from CONTAINER_DISCOVERY_MATRIX.values()


@pytest.mark.containers
@pytest.mark.parametrize(
    ('container', 'expected_package_manager'),
    list(_parametrize_test_discovery()),
    indirect=["container"],
    ids=[container.url for container, _ in CONTAINER_DISCOVERY_MATRIX.values()],
)
def test_discovery(
    container: ContainerData,
    guest: GuestContainer,
    expected_package_manager: PackageManagerClass,
    root_logger: tmt.log.Logger,
    caplog: _pytest.logging.LogCaptureFixture,
) -> None:
    guest.show()

    def _test_discovery(expected: str, expected_discovery: str) -> None:
        caplog.clear()

        guest.facts.sync(guest)

        assert guest.facts.package_manager == expected

        assert_log(caplog, message=MATCH(rf"^Discovered package managers: {expected_discovery}$"))

    # Images in which `dnf5`` would be the best possible choice, do not
    # come with `dnf5`` pre-installed. Therefore run the discovery first,
    # but expect to find *dnf* instead of `dnf5`. Then install `dnf5`,
    # re-run the discovery and expect the original outcome.
    if expected_package_manager is tmt.package_managers.dnf.Dnf5:
        guest.info(f'{container.image_url_or_id=}')
        guest.info(f'{has_legacy_dnf(container)=}')
        guest.info(f'{has_dnf5_preinstalled(container)=}')

        if has_legacy_dnf(container):
            _test_discovery(tmt.package_managers.dnf.Dnf.NAME, 'dnf and yum')

            if not has_dnf5_preinstalled(container):
                guest.execute(ShellScript('dnf install --nogpgcheck -y dnf5'))

        if has_legacy_dnf(container):
            _test_discovery(expected_package_manager.NAME, 'dnf5, dnf and yum')

        else:
            _test_discovery(expected_package_manager.NAME, 'dnf5')

    elif expected_package_manager is tmt.package_managers.dnf.Dnf:
        _test_discovery(expected_package_manager.NAME, 'dnf and yum')

    elif expected_package_manager is tmt.package_managers.rpm_ostree.RpmOstree:
        _test_discovery(expected_package_manager.NAME, 'rpm-ostree and dnf5')

    else:
        _test_discovery(expected_package_manager.NAME, expected_package_manager.NAME)


def _parametrize_test_install() -> Iterator[
    tuple[Container, PackageManagerClass, Package, str, Optional[str]]
]:
    for container, package_manager_class in CONTAINER_BASE_MATRIX:
        if package_manager_class is tmt.package_managers.dnf.Yum:
            if container.url == CONTAINER_FEDORA_RAWHIDE.url:
                yield (
                    container,
                    package_manager_class,
                    Package('tree'),
                    r"rpm -q --whatprovides tree \|\| yum install -y  tree && rpm -q --whatprovides tree",  # noqa: E501
                    'Installing:',
                )

            elif 'ubi/8' in container.url:
                yield (
                    container,
                    package_manager_class,
                    Package('dconf'),
                    r"rpm -q --whatprovides dconf \|\| yum install -y  dconf && rpm -q --whatprovides dconf",  # noqa: E501
                    'Installed:\n  dconf',
                )

            else:
                yield (
                    container,
                    package_manager_class,
                    Package('tree'),
                    r"rpm -q --whatprovides tree \|\| yum install -y  tree && rpm -q --whatprovides tree",  # noqa: E501
                    'Installed:\n  tree',
                )

        elif package_manager_class is tmt.package_managers.dnf.Dnf:
            if container.url == CONTAINER_FEDORA_RAWHIDE.url:
                yield (
                    container,
                    package_manager_class,
                    Package('tree'),
                    r"rpm -q --whatprovides tree \|\| dnf install -y  tree",
                    'Installing:',
                )

            elif 'ubi/8' in container.url:
                yield (
                    container,
                    package_manager_class,
                    Package('dconf'),
                    r"rpm -q --whatprovides dconf \|\| dnf install -y  dconf",
                    'Installed:\n  dconf',
                )

            else:
                yield (
                    container,
                    package_manager_class,
                    Package('tree'),
                    r"rpm -q --whatprovides tree \|\| dnf install -y  tree",
                    'Installed:\n  tree',
                )

        elif package_manager_class is tmt.package_managers.dnf.Dnf5:
            yield (
                container,
                package_manager_class,
                Package('tree'),
                r"rpm -q --whatprovides tree \|\| dnf5 install -y  tree",
                'Installing:',
            )

        elif package_manager_class is tmt.package_managers.apt.Apt:
            yield (
                container,
                package_manager_class,
                Package('tree'),
                r"set -x\s+export DEBIAN_FRONTEND=noninteractive\s+installable_packages=\"tree\"\s+dpkg-query --show \$installable_packages \\\s+\|\| apt install -y  \$installable_packages\s+exit \$\?",  # noqa: E501
                'Setting up tree',
            )

        elif package_manager_class is tmt.package_managers.rpm_ostree.RpmOstree:
            yield (
                container,
                package_manager_class,
                Package('tree'),
                r"rpm -q --whatprovides tree \|\| rpm-ostree install --apply-live --idempotent --allow-inactive --assumeyes  tree",  # noqa: E501
                'Installing: tree',
            )

        elif package_manager_class is tmt.package_managers.apk.Apk:
            yield (
                container,
                package_manager_class,
                Package('tree'),
                r"apk info -e tree \|\| apk add tree",
                'Installing tree',
            )

        else:
            pytest.fail(f"Unhandled package manager class '{package_manager_class}'.")


@pytest.mark.containers
@pytest.mark.parametrize(
    (
        'container_per_test',
        'package_manager_class',
        'package',
        'expected_command',
        'expected_output',
    ),
    list(_parametrize_test_install()),
    indirect=["container_per_test"],
    ids=CONTAINER_MATRIX_IDS,
)
def test_install(
    container_per_test: ContainerData,
    guest_per_test: GuestContainer,
    package_manager_class: PackageManagerClass,
    package: Package,
    expected_command: str,
    expected_output: Optional[str],
    root_logger: tmt.log.Logger,
    caplog: _pytest.logging.LogCaptureFixture,
) -> None:
    package_manager = create_package_manager(
        container_per_test, guest_per_test, package_manager_class, root_logger
    )

    output = package_manager.install(package)

    assert_log(
        caplog,
        message=MATCH(rf"(?sm)Run command: podman exec .+? /bin/bash -c '{expected_command}'"),
    )

    assert_output(expected_output, output.stdout, output.stderr)


def _parametrize_test_refresh_metadata() -> Iterator[
    tuple[Container, PackageManagerClass, str, Optional[str]]
]:
    for container, package_manager_class in CONTAINER_BASE_MATRIX:
        if package_manager_class is tmt.package_managers.dnf.Yum:
            yield container, package_manager_class, r"yum makecache", 'Metadata'

        elif package_manager_class is tmt.package_managers.dnf.Dnf:
            yield (
                container,
                package_manager_class,
                r"dnf makecache -y --refresh",
                'Metadata cache created',
            )

        elif package_manager_class is tmt.package_managers.dnf.Dnf5:
            yield (
                container,
                package_manager_class,
                r"dnf5 makecache -y --refresh",
                'Metadata cache created',
            )

        elif package_manager_class is tmt.package_managers.apt.Apt:
            yield (
                container,
                package_manager_class,
                r"export DEBIAN_FRONTEND=noninteractive; apt update",
                'Reading package list',
            )

        elif package_manager_class is tmt.package_managers.rpm_ostree.RpmOstree:
            yield pytest.param(
                container,
                package_manager_class,
                r"rpm-ostree refresh-md --force",
                'Available',
                marks=pytest.mark.skip(
                    reason="refresh-md does not work with how tmt runs ostree container"
                ),
            )

        elif package_manager_class is tmt.package_managers.apk.Apk:
            yield container, package_manager_class, r"apk update", 'OK:'

        else:
            pytest.fail(f"Unhandled package manager class '{package_manager_class}'.")


@pytest.mark.containers
@pytest.mark.parametrize(
    ('container_per_test', 'package_manager_class', 'expected_command', 'expected_output'),
    list(_parametrize_test_refresh_metadata()),
    indirect=["container_per_test"],
    ids=CONTAINER_MATRIX_IDS,
)
def test_refresh_metadata(
    container_per_test: ContainerData,
    guest_per_test: GuestContainer,
    package_manager_class: PackageManagerClass,
    expected_command: str,
    expected_output: Optional[str],
    root_logger: tmt.log.Logger,
    caplog: _pytest.logging.LogCaptureFixture,
) -> None:
    package_manager = create_package_manager(
        container_per_test, guest_per_test, package_manager_class, root_logger
    )

    output = package_manager.refresh_metadata()

    assert_log(
        caplog, message=MATCH(rf"Run command: podman exec .+? /bin/bash -c '{expected_command}'")
    )

    assert_output(expected_output, output.stdout, output.stderr)


def _parametrize_test_install_nonexistent() -> Iterator[
    tuple[Container, PackageManagerClass, str, Optional[str]]
]:
    for container, package_manager_class in CONTAINER_BASE_MATRIX:
        if package_manager_class is tmt.package_managers.dnf.Dnf5:
            yield (
                container,
                package_manager_class,
                r"rpm -q --whatprovides tree-but-spelled-wrong \|\| dnf5 install -y  tree-but-spelled-wrong",  # noqa: E501
                'No match for argument: tree-but-spelled-wrong',
            )

        elif package_manager_class is tmt.package_managers.dnf.Dnf:
            if container.url == CONTAINER_FEDORA_RAWHIDE.url:
                yield (
                    container,
                    package_manager_class,
                    r"rpm -q --whatprovides tree-but-spelled-wrong \|\| dnf install -y  tree-but-spelled-wrong",  # noqa: E501
                    'No match for argument: tree-but-spelled-wrong',
                )

            else:
                yield (
                    container,
                    package_manager_class,
                    r"rpm -q --whatprovides tree-but-spelled-wrong \|\| dnf install -y  tree-but-spelled-wrong",  # noqa: E501
                    'Error: Unable to find a match: tree-but-spelled-wrong',
                )

        elif package_manager_class is tmt.package_managers.dnf.Yum:
            if container.url == CONTAINER_FEDORA_RAWHIDE.url:
                yield (
                    container,
                    package_manager_class,
                    r"rpm -q --whatprovides tree-but-spelled-wrong \|\| yum install -y  tree-but-spelled-wrong && rpm -q --whatprovides tree-but-spelled-wrong",  # noqa: E501
                    'No match for argument: tree-but-spelled-wrong',
                )

            elif 'fedora' in container.url:
                yield (
                    container,
                    package_manager_class,
                    r"rpm -q --whatprovides tree-but-spelled-wrong \|\| yum install -y  tree-but-spelled-wrong && rpm -q --whatprovides tree-but-spelled-wrong",  # noqa: E501
                    'Error: Unable to find a match: tree-but-spelled-wrong',
                )

            elif (
                'centos' in container.url and 'centos/7' not in container.url
            ) or 'ubi/8' in container.url:
                yield (
                    container,
                    package_manager_class,
                    r"rpm -q --whatprovides tree-but-spelled-wrong \|\| yum install -y  tree-but-spelled-wrong && rpm -q --whatprovides tree-but-spelled-wrong",  # noqa: E501
                    'No match for argument: tree-but-spelled-wrong',
                )

            else:
                yield (
                    container,
                    package_manager_class,
                    r"rpm -q --whatprovides tree-but-spelled-wrong \|\| yum install -y  tree-but-spelled-wrong && rpm -q --whatprovides tree-but-spelled-wrong",  # noqa: E501
                    'No package tree-but-spelled-wrong available.',
                )

        elif package_manager_class is tmt.package_managers.apt.Apt:
            yield (
                container,
                package_manager_class,
                r"set -x\s+export DEBIAN_FRONTEND=noninteractive\s+installable_packages=\"tree-but-spelled-wrong\"\s+dpkg-query --show \$installable_packages \\\s+\|\| apt install -y  \$installable_packages\s+exit \$\?",  # noqa: E501
                'E: Unable to locate package tree-but-spelled-wrong',
            )

        elif package_manager_class is tmt.package_managers.rpm_ostree.RpmOstree:
            yield (
                container,
                package_manager_class,
                r"rpm -q --whatprovides tree-but-spelled-wrong \|\| rpm-ostree install --apply-live --idempotent --allow-inactive --assumeyes  tree-but-spelled-wrong",  # noqa: E501
                'no package provides tree-but-spelled-wrong',
            )

        elif package_manager_class is tmt.package_managers.apk.Apk:
            yield (
                container,
                package_manager_class,
                r"apk info -e tree-but-spelled-wrong \|\| apk add tree-but-spelled-wrong",
                'ERROR: unable to select packages:\n  tree-but-spelled-wrong (no such package):\n    required by: world[tree-but-spelled-wrong]',  # noqa: E501
            )

        else:
            pytest.fail(f"Unhandled package manager class '{package_manager_class}'.")


@pytest.mark.containers
@pytest.mark.parametrize(
    ('container', 'package_manager_class', 'expected_command', 'expected_output'),
    list(_parametrize_test_install_nonexistent()),
    indirect=["container"],
    ids=CONTAINER_MATRIX_IDS,
)
def test_install_nonexistent(
    container: ContainerData,
    guest: GuestContainer,
    package_manager_class: PackageManagerClass,
    expected_command: str,
    expected_output: Optional[str],
    root_logger: tmt.log.Logger,
    caplog: _pytest.logging.LogCaptureFixture,
) -> None:
    package_manager = create_package_manager(container, guest, package_manager_class, root_logger)

    with pytest.raises(tmt.utils.RunError) as excinfo:
        package_manager.install(Package('tree-but-spelled-wrong'))

    assert_log(
        caplog,
        message=MATCH(rf"(?sm)Run command: podman exec .+? /bin/bash -c '{expected_command}'"),
    )

    assert excinfo.type is tmt.utils.RunError
    assert excinfo.value.returncode != 0

    assert_output(expected_output, excinfo.value.stdout, excinfo.value.stderr)


def _parametrize_test_install_nonexistent_skip() -> Iterator[
    tuple[Container, PackageManagerClass, str, Optional[str]]
]:
    for container, package_manager_class in CONTAINER_BASE_MATRIX:
        if package_manager_class is tmt.package_managers.dnf.Dnf5:
            yield (
                container,
                package_manager_class,
                r"rpm -q --whatprovides tree-but-spelled-wrong \|\| dnf5 install -y --skip-unavailable tree-but-spelled-wrong",  # noqa: E501
                None,
            )

        elif package_manager_class is tmt.package_managers.dnf.Dnf:
            yield (
                container,
                package_manager_class,
                r"rpm -q --whatprovides tree-but-spelled-wrong \|\| dnf install -y --skip-broken tree-but-spelled-wrong",  # noqa: E501
                'No match for argument: tree-but-spelled-wrong',
            )

        elif package_manager_class is tmt.package_managers.dnf.Yum:
            if container.url == CONTAINER_FEDORA_RAWHIDE.url:  # noqa: SIM114
                yield (
                    container,
                    package_manager_class,
                    r"rpm -q --whatprovides tree-but-spelled-wrong \|\| yum install -y --skip-broken tree-but-spelled-wrong \|\| /bin/true",  # noqa: E501
                    'No match for argument: tree-but-spelled-wrong',
                )

            elif 'fedora' in container.url:  # noqa: SIM114
                yield (
                    container,
                    package_manager_class,
                    r"rpm -q --whatprovides tree-but-spelled-wrong \|\| yum install -y --skip-broken tree-but-spelled-wrong \|\| /bin/true",  # noqa: E501
                    'No match for argument: tree-but-spelled-wrong',
                )

            elif (
                'centos' in container.url and 'centos/7' not in container.url
            ) or 'ubi/8' in container.url:
                yield (
                    container,
                    package_manager_class,
                    r"rpm -q --whatprovides tree-but-spelled-wrong \|\| yum install -y --skip-broken tree-but-spelled-wrong \|\| /bin/true",  # noqa: E501
                    'No match for argument: tree-but-spelled-wrong',
                )

            else:
                yield (
                    container,
                    package_manager_class,
                    r"rpm -q --whatprovides tree-but-spelled-wrong \|\| yum install -y --skip-broken tree-but-spelled-wrong \|\| /bin/true",  # noqa: E501
                    'No package tree-but-spelled-wrong available.',
                )

        elif package_manager_class is tmt.package_managers.apt.Apt:
            yield (
                container,
                package_manager_class,
                r"set -x\s+export DEBIAN_FRONTEND=noninteractive\s+installable_packages=\"tree-but-spelled-wrong\"\s+dpkg-query --show \$installable_packages \\\s+\|\| apt install -y --ignore-missing \$installable_packages\s+exit 0",  # noqa: E501
                'E: Unable to locate package tree-but-spelled-wrong',
            )

        elif package_manager_class is tmt.package_managers.rpm_ostree.RpmOstree:
            yield (
                container,
                package_manager_class,
                r"rpm -q --whatprovides tree-but-spelled-wrong \|\| rpm-ostree install --apply-live --idempotent --allow-inactive --assumeyes  tree-but-spelled-wrong \|\| /bin/true",  # noqa: E501
                'no package provides tree-but-spelled-wrong',
            )

        elif package_manager_class is tmt.package_managers.apk.Apk:
            yield (
                container,
                package_manager_class,
                r"apk info -e tree-but-spelled-wrong \|\| apk add tree-but-spelled-wrong \|\| /bin/true",  # noqa: E501
                'ERROR: unable to select packages:\n  tree-but-spelled-wrong (no such package):\n    required by: world[tree-but-spelled-wrong]',  # noqa: E501
            )

        else:
            pytest.fail(f"Unhandled package manager class '{package_manager_class}'.")


@pytest.mark.containers
@pytest.mark.parametrize(
    ('container', 'package_manager_class', 'expected_command', 'expected_output'),
    list(_parametrize_test_install_nonexistent_skip()),
    indirect=["container"],
    ids=CONTAINER_MATRIX_IDS,
)
def test_install_nonexistent_skip(
    container: ContainerData,
    guest: GuestContainer,
    package_manager_class: PackageManagerClass,
    expected_command: str,
    expected_output: Optional[str],
    root_logger: tmt.log.Logger,
    caplog: _pytest.logging.LogCaptureFixture,
) -> None:
    package_manager = create_package_manager(container, guest, package_manager_class, root_logger)

    output = package_manager.install(
        Package('tree-but-spelled-wrong'), options=Options(skip_missing=True)
    )

    assert_log(
        caplog,
        message=MATCH(rf"(?sm)Run command: podman exec .+? /bin/bash -c '{expected_command}'"),
    )

    assert_output(expected_output, output.stdout, output.stderr)


def _parametrize_test_install_dont_check_first() -> Iterator[
    tuple[Container, PackageManagerClass, Package, str, Optional[str]]
]:
    for container, package_manager_class in CONTAINER_BASE_MATRIX:
        if package_manager_class is tmt.package_managers.dnf.Dnf5:
            yield container, package_manager_class, Package('tree'), r"dnf5 install -y  tree", None

        elif package_manager_class is tmt.package_managers.dnf.Dnf:
            if 'ubi/8' in container.url:
                yield (
                    container,
                    package_manager_class,
                    Package('dconf'),
                    r"dnf install -y  dconf",
                    'Installed:\n  dconf',
                )
            else:
                yield (
                    container,
                    package_manager_class,
                    Package('tree'),
                    r"dnf install -y  tree",
                    'Installed:\n  tree',
                )

        elif package_manager_class is tmt.package_managers.dnf.Yum:
            if 'ubi/8' in container.url:
                yield (
                    container,
                    package_manager_class,
                    Package('dconf'),
                    r"yum install -y  dconf && rpm -q --whatprovides dconf",
                    'Installed:\n  dconf',
                )

            else:
                yield (
                    container,
                    package_manager_class,
                    Package('tree'),
                    r"yum install -y  tree && rpm -q --whatprovides tree",
                    'Installed:\n  tree',
                )

        elif package_manager_class is tmt.package_managers.apt.Apt:
            yield (
                container,
                package_manager_class,
                Package('tree'),
                r"set -x\s+export DEBIAN_FRONTEND=noninteractive\s+installable_packages=\"tree\"\s+/bin/false \\\s+\|\| apt install -y  \$installable_packages\s+exit \$\?",  # noqa: E501
                'Setting up tree',
            )

        elif package_manager_class is tmt.package_managers.rpm_ostree.RpmOstree:
            yield (
                container,
                package_manager_class,
                Package('tree'),
                r"rpm-ostree install --apply-live --idempotent --allow-inactive --assumeyes  tree",
                'Installing: tree',
            )

        elif package_manager_class is tmt.package_managers.apk.Apk:
            yield (
                container,
                package_manager_class,
                Package('tree'),
                r"apk add tree",
                'Installing tree',
            )

        else:
            pytest.fail(f"Unhandled package manager class '{package_manager_class}'.")


@pytest.mark.containers
@pytest.mark.parametrize(
    (
        'container_per_test',
        'package_manager_class',
        'package',
        'expected_command',
        'expected_output',
    ),
    list(_parametrize_test_install_dont_check_first()),
    indirect=["container_per_test"],
    ids=CONTAINER_MATRIX_IDS,
)
def test_install_dont_check_first(
    container_per_test: ContainerData,
    guest_per_test: GuestContainer,
    package_manager_class: PackageManagerClass,
    package: Package,
    expected_command: str,
    expected_output: Optional[str],
    root_logger: tmt.log.Logger,
    caplog: _pytest.logging.LogCaptureFixture,
) -> None:
    package_manager = create_package_manager(
        container_per_test, guest_per_test, package_manager_class, root_logger
    )

    output = package_manager.install(package, options=Options(check_first=False))

    assert_log(
        caplog,
        message=MATCH(rf"(?sm)Run command: podman exec .+? /bin/bash -c '{expected_command}'"),
    )

    assert_output(expected_output, output.stdout, output.stderr)


def _parametrize_test_reinstall() -> Iterator[
    tuple[Container, PackageManagerClass, Optional[str], Optional[str]]
]:
    for container, package_manager_class in CONTAINER_BASE_MATRIX:
        if package_manager_class is tmt.package_managers.dnf.Yum:
            if 'centos/7' in container.url:
                yield (
                    container,
                    package_manager_class,
                    Package('tar'),
                    True,
                    r"rpm -q --whatprovides tar && yum reinstall -y  tar && rpm -q --whatprovides tar",  # noqa: E501
                    'Reinstalling:\n tar',
                )

            else:
                yield (
                    container,
                    package_manager_class,
                    Package('tar'),
                    True,
                    r"rpm -q --whatprovides tar && yum reinstall -y  tar && rpm -q --whatprovides tar",  # noqa: E501
                    'Reinstalled:\n  tar',
                )

        elif package_manager_class is tmt.package_managers.dnf.Dnf:
            yield (
                container,
                package_manager_class,
                Package('tar'),
                True,
                r"rpm -q --whatprovides tar && dnf reinstall -y  tar",
                'Reinstalled:\n  tar',
            )

        elif package_manager_class is tmt.package_managers.dnf.Dnf5:
            yield (
                container,
                package_manager_class,
                Package('tar'),
                True,
                r"rpm -q --whatprovides tar && dnf5 reinstall -y  tar",
                'Reinstalling tar',
            )

        elif package_manager_class is tmt.package_managers.apt.Apt:
            yield (
                container,
                package_manager_class,
                Package('tar'),
                True,
                r"set -x\s+export DEBIAN_FRONTEND=noninteractive\s+installable_packages=\"tar\"\s+dpkg-query --show \$installable_packages \\\s+&& apt reinstall -y  \$installable_packages\s+exit \$\?",  # noqa: E501
                'Setting up tar',
            )

        elif package_manager_class is tmt.package_managers.rpm_ostree.RpmOstree:
            yield container, package_manager_class, Package('tar'), False, None, None

        elif package_manager_class is tmt.package_managers.apk.Apk:
            yield (
                container,
                package_manager_class,
                Package('bash'),
                True,
                r"apk info -e bash && apk fix bash",
                'Reinstalling bash',
            )

        else:
            pytest.fail(f"Unhandled package manager class '{package_manager_class}'.")


@pytest.mark.containers
@pytest.mark.parametrize(
    (
        'container_per_test',
        'package_manager_class',
        'package',
        'supported',
        'expected_command',
        'expected_output',
    ),
    list(_parametrize_test_reinstall()),
    indirect=["container_per_test"],
    ids=CONTAINER_MATRIX_IDS,
)
def test_reinstall(
    container_per_test: ContainerData,
    guest_per_test: GuestContainer,
    package_manager_class: PackageManagerClass,
    package: Package,
    supported: bool,
    expected_command: Optional[str],
    expected_output: Optional[str],
    root_logger: tmt.log.Logger,
    caplog: _pytest.logging.LogCaptureFixture,
) -> None:
    package_manager = create_package_manager(
        container_per_test, guest_per_test, package_manager_class, root_logger
    )

    if supported:
        assert expected_command is not None

        output = package_manager.reinstall(package)

        assert_log(
            caplog,
            message=MATCH(rf"(?sm)Run command: podman exec .+? /bin/bash -c '{expected_command}'"),
        )

    else:
        with pytest.raises(tmt.utils.GeneralError) as excinfo:
            package_manager.reinstall(package)

        assert excinfo.value.message == "rpm-ostree does not support reinstall operation."

    if expected_output:
        assert_output(expected_output, output.stdout, output.stderr)


def _generate_test_reinstall_nonexistent_matrix() -> Iterator[
    tuple[Container, PackageManagerClass, Optional[str], Optional[str], Optional[str]]
]:
    for container, package_manager_class in CONTAINER_BASE_MATRIX:
        if package_manager_class is tmt.package_managers.dnf.Dnf5:
            yield (
                container,
                package_manager_class,
                True,
                r"rpm -q --whatprovides tree-but-spelled-wrong && dnf5 reinstall -y  tree-but-spelled-wrong",  # noqa: E501
                'no package provides tree-but-spelled-wrong',
            )

        elif package_manager_class is tmt.package_managers.dnf.Dnf:
            yield (
                container,
                package_manager_class,
                True,
                r"rpm -q --whatprovides tree-but-spelled-wrong && dnf reinstall -y  tree-but-spelled-wrong",  # noqa: E501
                'no package provides tree-but-spelled-wrong',
            )

        elif package_manager_class is tmt.package_managers.dnf.Yum:
            if 'fedora' in container.url:  # noqa: SIM114
                yield (
                    container,
                    package_manager_class,
                    True,
                    r"rpm -q --whatprovides tree-but-spelled-wrong && yum reinstall -y  tree-but-spelled-wrong && rpm -q --whatprovides tree-but-spelled-wrong",  # noqa: E501
                    'no package provides tree-but-spelled-wrong',
                )

            elif 'centos' in container.url and 'centos/7' not in container.url:
                yield (
                    container,
                    package_manager_class,
                    True,
                    r"rpm -q --whatprovides tree-but-spelled-wrong && yum reinstall -y  tree-but-spelled-wrong && rpm -q --whatprovides tree-but-spelled-wrong",  # noqa: E501
                    'no package provides tree-but-spelled-wrong',
                )

            else:
                yield (
                    container,
                    package_manager_class,
                    True,
                    r"rpm -q --whatprovides tree-but-spelled-wrong && yum reinstall -y  tree-but-spelled-wrong && rpm -q --whatprovides tree-but-spelled-wrong",  # noqa: E501
                    'no package provides tree-but-spelled-wrong',
                )

        elif package_manager_class is tmt.package_managers.apt.Apt:
            yield (
                container,
                package_manager_class,
                True,
                r"set -x\s+export DEBIAN_FRONTEND=noninteractive\s+installable_packages=\"tree-but-spelled-wrong\"\s+dpkg-query --show \$installable_packages \\\s+&& apt reinstall -y  \$installable_packages\s+exit \$\?",  # noqa: E501
                'dpkg-query: no packages found matching tree-but-spelled-wrong',
            )

        elif package_manager_class is tmt.package_managers.rpm_ostree.RpmOstree:
            yield container, package_manager_class, False, None, None

        elif package_manager_class is tmt.package_managers.apk.Apk:
            yield (
                container,
                package_manager_class,
                True,
                r"apk info -e tree-but-spelled-wrong && apk fix tree-but-spelled-wrong",
                None,
            )

        else:
            pytest.fail(f"Unhandled package manager class '{package_manager_class}'.")


@pytest.mark.containers
@pytest.mark.parametrize(
    ('container', 'package_manager_class', 'supported', 'expected_command', 'expected_output'),
    list(_generate_test_reinstall_nonexistent_matrix()),
    indirect=["container"],
    ids=CONTAINER_MATRIX_IDS,
)
def test_reinstall_nonexistent(
    container: ContainerData,
    guest: GuestContainer,
    package_manager_class: PackageManagerClass,
    supported: bool,
    expected_command: str,
    expected_output: Optional[str],
    root_logger: tmt.log.Logger,
    caplog: _pytest.logging.LogCaptureFixture,
) -> None:
    package_manager = create_package_manager(container, guest, package_manager_class, root_logger)

    if supported:
        assert expected_command is not None

        with pytest.raises(tmt.utils.RunError) as excinfo:
            package_manager.reinstall(Package('tree-but-spelled-wrong'))

        assert_log(
            caplog,
            message=MATCH(rf"(?sm)Run command: podman exec .+? /bin/bash -c '{expected_command}'"),
        )

        assert excinfo.type is tmt.utils.RunError
        assert excinfo.value.returncode != 0

    else:
        with pytest.raises(tmt.utils.GeneralError) as excinfo:
            package_manager.reinstall(Package('tree-but-spelled-wrong'))

        assert excinfo.value.message == "rpm-ostree does not support reinstall operation."

    if expected_output:
        assert_output(expected_output, excinfo.value.stdout, excinfo.value.stderr)


def _generate_test_check_presence() -> Iterator[
    tuple[Container, PackageManagerClass, Installable, str, Optional[str]]
]:
    for container, package_manager_class in CONTAINER_BASE_MATRIX:
        if package_manager_class is tmt.package_managers.dnf.Dnf5:
            yield (
                container,
                package_manager_class,
                Package('coreutils'),
                True,
                r"rpm -q --whatprovides coreutils",
                r'\s+out:\s+coreutils-',
            )

            yield (
                container,
                package_manager_class,
                Package('tree-but-spelled-wrong'),
                False,
                r"rpm -q --whatprovides tree-but-spelled-wrong",
                r'\s+out:\s+no package provides tree-but-spelled-wrong',
            )

            yield (
                container,
                package_manager_class,
                FileSystemPath('/usr/bin/arch'),
                True,
                r"rpm -q --whatprovides /usr/bin/arch",
                r'\s+out:\s+coreutils-',
            )

        elif package_manager_class is tmt.package_managers.dnf.Dnf:
            if 'ubi/8' in container.url:
                yield (
                    container,
                    package_manager_class,
                    Package('util-linux'),
                    True,
                    r"rpm -q --whatprovides util-linux",
                    r'\s+out:\s+util-linux-',
                )

                yield (
                    container,
                    package_manager_class,
                    Package('tree-but-spelled-wrong'),
                    False,
                    r"rpm -q --whatprovides tree-but-spelled-wrong",
                    r'\s+out:\s+no package provides tree-but-spelled-wrong',
                )

                yield (
                    container,
                    package_manager_class,
                    FileSystemPath('/usr/bin/flock'),
                    True,
                    r"rpm -q --whatprovides /usr/bin/flock",
                    r'\s+out:\s+util-linux-',
                )

            elif 'centos/stream9' in container.url or 'fedora/40' in container.url:
                yield (
                    container,
                    package_manager_class,
                    Package('coreutils'),
                    True,
                    r"rpm -q --whatprovides coreutils",
                    r'\s+out:\s+coreutils-',
                )

                yield (
                    container,
                    package_manager_class,
                    Package('tree-but-spelled-wrong'),
                    False,
                    r"rpm -q --whatprovides tree-but-spelled-wrong",
                    r'\s+out:\s+no package provides tree-but-spelled-wrong',
                )

                yield (
                    container,
                    package_manager_class,
                    FileSystemPath('/usr/bin/arch'),
                    True,
                    r"rpm -q --whatprovides /usr/bin/arch",
                    r'\s+out:\s+coreutils-',
                )

            else:
                yield (
                    container,
                    package_manager_class,
                    Package('util-linux-core'),
                    True,
                    r"rpm -q --whatprovides util-linux-core",
                    r'\s+out:\s+util-linux-core-',
                )

                yield (
                    container,
                    package_manager_class,
                    Package('tree-but-spelled-wrong'),
                    False,
                    r"rpm -q --whatprovides tree-but-spelled-wrong",
                    r'\s+out:\s+no package provides tree-but-spelled-wrong',
                )

                yield (
                    container,
                    package_manager_class,
                    FileSystemPath('/usr/bin/flock'),
                    True,
                    r"rpm -q --whatprovides /usr/bin/flock",
                    r'\s+out:\s+util-linux-core-',
                )

        elif package_manager_class is tmt.package_managers.dnf.Yum:
            if 'centos/7' in container.url or 'ubi/8' in container.url:
                yield (
                    container,
                    package_manager_class,
                    Package('util-linux'),
                    True,
                    r"rpm -q --whatprovides util-linux",
                    r'\s+out:\s+util-linux-',
                )

                yield (
                    container,
                    package_manager_class,
                    Package('tree-but-spelled-wrong'),
                    False,
                    r"rpm -q --whatprovides tree-but-spelled-wrong",
                    r'\s+out:\s+no package provides tree-but-spelled-wrong',
                )

                yield (
                    container,
                    package_manager_class,
                    FileSystemPath('/usr/bin/flock'),
                    True,
                    r"rpm -q --whatprovides /usr/bin/flock",
                    r'\s+out:\s+util-linux-',
                )

            elif 'centos/stream9' in container.url or 'fedora/40' in container.url:
                yield (
                    container,
                    package_manager_class,
                    Package('coreutils'),
                    True,
                    r"rpm -q --whatprovides coreutils",
                    r'\s+out:\s+coreutils-',
                )

                yield (
                    container,
                    package_manager_class,
                    Package('tree-but-spelled-wrong'),
                    False,
                    r"rpm -q --whatprovides tree-but-spelled-wrong",
                    r'\s+out:\s+no package provides tree-but-spelled-wrong',
                )

                yield (
                    container,
                    package_manager_class,
                    FileSystemPath('/usr/bin/arch'),
                    True,
                    r"rpm -q --whatprovides /usr/bin/arch",
                    r'\s+out:\s+coreutils-',
                )

            else:
                yield (
                    container,
                    package_manager_class,
                    Package('util-linux-core'),
                    True,
                    r"rpm -q --whatprovides util-linux-core",
                    r'\s+out:\s+util-linux-core-',
                )

                yield (
                    container,
                    package_manager_class,
                    Package('tree-but-spelled-wrong'),
                    False,
                    r"rpm -q --whatprovides tree-but-spelled-wrong",
                    r'\s+out:\s+no package provides tree-but-spelled-wrong',
                )

                yield (
                    container,
                    package_manager_class,
                    FileSystemPath('/usr/bin/flock'),
                    True,
                    r"rpm -q --whatprovides /usr/bin/flock",
                    r'\s+out:\s+util-linux-core-',
                )

        elif package_manager_class is tmt.package_managers.apt.Apt:
            yield (
                container,
                package_manager_class,
                Package('util-linux'),
                True,
                r".*?echo \"PRESENCE-TEST:util-linux:util-linux:\$\(dpkg-query --show util-linux\)\".*?",  # noqa: E501
                r'\s+out:\s+PRESENCE-TEST:util-linux:util-linux:util-linux',
            )

            yield (
                container,
                package_manager_class,
                Package('tree-but-spelled-wrong'),
                False,
                r".*?echo \"PRESENCE-TEST:tree-but-spelled-wrong:tree-but-spelled-wrong:\$\(dpkg-query --show tree-but-spelled-wrong\)\".*?",  # noqa: E501
                r'\s+err:\s+dpkg-query: no packages found matching tree-but-spelled-wrong',
            )

            yield (
                container,
                package_manager_class,
                FileSystemPath('/usr/bin/flock'),
                True,
                r".*?set -x\s+export DEBIAN_FRONTEND=noninteractive\s+fs_path_package=\"\$\(apt-file search --package-only /usr/bin/flock\)\".*echo \"PRESENCE-TEST:/usr/bin/flock:\$\{fs_path_package\}:\$\(dpkg-query --show \$fs_path_package\)\".*?",  # noqa: E501
                r'\s+out:\s+PRESENCE-TEST:/usr/bin/flock:util-linux:util-linux',
            )

        elif package_manager_class is tmt.package_managers.rpm_ostree.RpmOstree:
            yield (
                container,
                package_manager_class,
                Package('util-linux'),
                True,
                r"rpm -q --whatprovides util-linux",
                r'\s+out:\s+util-linux',
            )

            yield (
                container,
                package_manager_class,
                Package('tree-but-spelled-wrong'),
                False,
                r"rpm -q --whatprovides tree-but-spelled-wrong",
                r'\s+out:\s+no package provides tree-but-spelled-wrong',
            )

            yield (
                container,
                package_manager_class,
                FileSystemPath('/usr/bin/flock'),
                True,
                r"rpm -qf /usr/bin/flock",
                r'\s+out:\s+util-linux-core',
            )

        elif package_manager_class is tmt.package_managers.apk.Apk:
            yield (
                container,
                package_manager_class,
                Package('busybox'),
                True,
                r"apk info -e busybox",
                r'\s+out:\s+busybox',
            )

            yield (
                container,
                package_manager_class,
                Package('tree-but-spelled-wrong'),
                False,
                r"apk info -e tree-but-spelled-wrong",
                None,
            )

            yield (
                container,
                package_manager_class,
                FileSystemPath('/usr/bin/arch'),
                True,
                r"apk info -e busybox",
                r'\s+out:\s+busybox',
            )

        else:
            pytest.fail(f"Unhandled package manager class '{package_manager_class}'.")


def _generate_test_check_presence_ids(value) -> str:
    if isinstance(value, Container):
        return value.url

    if isclass(value) and issubclass(value, tmt.package_managers.PackageManager):
        return value.__name__.lower()

    if isinstance(value, (Package, FileSystemPath)):
        return str(value)

    return ''


@pytest.mark.containers
@pytest.mark.parametrize(
    (
        'container',
        'package_manager_class',
        'installable',
        'expected_result',
        'expected_command',
        'expected_output',
    ),
    list(_generate_test_check_presence()),
    indirect=["container"],
    ids=_generate_test_check_presence_ids,
)
def test_check_presence(
    container: ContainerData,
    guest: GuestContainer,
    package_manager_class: PackageManagerClass,
    installable: Installable,
    expected_result: bool,
    expected_command: str,
    expected_output: Optional[str],
    root_logger: tmt.log.Logger,
    caplog: _pytest.logging.LogCaptureFixture,
) -> None:
    package_manager = create_package_manager(container, guest, package_manager_class, root_logger)

    assert package_manager.check_presence(installable) == {installable: expected_result}

    assert_log(
        caplog,
        message=MATCH(rf"(?sm)Run command: podman exec .+? /bin/bash -c '{expected_command}'"),
    )

    if expected_output:
        assert_log(caplog, remove_colors=True, message=MATCH(expected_output))


def _parametrize_test_install_filesystempath() -> Iterator[
    tuple[Container, PackageManagerClass, FileSystemPath, Optional[str]]
]:
    for container, package_manager_class in CONTAINER_BASE_MATRIX:
        if package_manager_class is tmt.package_managers.dnf.Dnf5:
            yield (
                container,
                package_manager_class,
                FileSystemPath('/usr/bin/dos2unix'),
                r"rpm -q --whatprovides /usr/bin/dos2unix \|\| dnf5 install -y  /usr/bin/dos2unix",
                '[1/1] dos2unix',
            )

        elif package_manager_class is tmt.package_managers.dnf.Dnf:
            yield (
                container,
                package_manager_class,
                FileSystemPath('/usr/bin/dos2unix'),
                r"rpm -q --whatprovides /usr/bin/dos2unix \|\| dnf install -y  /usr/bin/dos2unix",
                'Installed:\n  dos2unix-',
            )

        elif package_manager_class is tmt.package_managers.dnf.Yum:
            if 'centos/7' in container.url:
                yield (
                    container,
                    package_manager_class,
                    FileSystemPath('/usr/bin/dos2unix'),
                    r"rpm -q --whatprovides /usr/bin/dos2unix \|\| yum install -y  /usr/bin/dos2unix && rpm -q --whatprovides /usr/bin/dos2unix",  # noqa: E501
                    'Installed:\n  dos2unix.',
                )

            else:
                yield (
                    container,
                    package_manager_class,
                    FileSystemPath('/usr/bin/dos2unix'),
                    r"rpm -q --whatprovides /usr/bin/dos2unix \|\| yum install -y  /usr/bin/dos2unix && rpm -q --whatprovides /usr/bin/dos2unix",  # noqa: E501
                    'Installed:\n  dos2unix-',
                )

        elif package_manager_class is tmt.package_managers.apt.Apt:
            yield (
                container,
                package_manager_class,
                FileSystemPath('/usr/bin/dos2unix'),
                r".*?set -x\s+export DEBIAN_FRONTEND=noninteractive\s+installable_packages=\"\"\s+fs_path_package=\"\$\(apt-file search --package-only /usr/bin/dos2unix\)\"\s+\[ -z \"\$fs_path_package\" \] && echo \"No package found for path /usr/bin/dos2unix\" && exit 1\s+installable_packages=\"\$installable_packages \$fs_path_package\"\s+dpkg-query --show \$installable_packages \\\s+\|\| apt install -y\s+\$installable_packages\s+exit \$\?",  # noqa: E501
                "Setting up dos2unix",
            )

        elif package_manager_class is tmt.package_managers.rpm_ostree.RpmOstree:
            yield (
                container,
                package_manager_class,
                FileSystemPath('/usr/bin/dos2unix'),
                r"rpm -qf /usr/bin/dos2unix \|\| rpm-ostree install --apply-live --idempotent --allow-inactive --assumeyes  /usr/bin/dos2unix",  # noqa: E501
                "Installing 1 packages:\n  dos2unix-",
            )

        elif package_manager_class is tmt.package_managers.apk.Apk:
            yield (
                container,
                package_manager_class,
                FileSystemPath('/usr/bin/dos2unix'),
                r"apk info -e dos2unix \|\| apk add dos2unix",
                'Installing dos2unix',
            )

        else:
            pytest.fail(f"Unhandled package manager class '{package_manager_class}'.")


@pytest.mark.containers
@pytest.mark.parametrize(
    (
        'container_per_test',
        'package_manager_class',
        'installable',
        'expected_command',
        'expected_output',
    ),
    list(_parametrize_test_install_filesystempath()),
    indirect=["container_per_test"],
    ids=CONTAINER_MATRIX_IDS,
)
def test_install_filesystempath(
    container_per_test: ContainerData,
    guest_per_test: GuestContainer,
    package_manager_class: PackageManagerClass,
    installable: FileSystemPath,
    expected_command: str,
    expected_output: Optional[str],
    root_logger: tmt.log.Logger,
    caplog: _pytest.logging.LogCaptureFixture,
) -> None:
    package_manager = create_package_manager(
        container_per_test, guest_per_test, package_manager_class, root_logger
    )

    output = package_manager.install(installable)

    assert_log(
        caplog,
        message=MATCH(rf"(?sm)Run command: podman exec .+? /bin/bash -c '{expected_command}'"),
    )

    assert_output(expected_output, output.stdout, output.stderr)


def _parametrize_test_install_multiple() -> Iterator[
    tuple[Container, PackageManagerClass, tuple[Package, Package], str, Optional[str]]
]:
    for container, package_manager_class in CONTAINER_BASE_MATRIX:
        if package_manager_class is tmt.package_managers.dnf.Yum:
            if 'ubi/8' in container.url:
                yield (
                    container,
                    package_manager_class,
                    (Package('dconf'), Package('libpng')),
                    r"rpm -q --whatprovides dconf libpng \|\| yum install -y  dconf libpng && rpm -q --whatprovides dconf libpng",  # noqa: E501
                    'Complete!',
                )

            elif 'centos' in container.url:
                yield (
                    container,
                    package_manager_class,
                    (Package('tree'), Package('diffutils')),
                    r"rpm -q --whatprovides tree diffutils \|\| yum install -y  tree diffutils && rpm -q --whatprovides tree diffutils",  # noqa: E501
                    'Complete!',
                )

            else:
                yield (
                    container,
                    package_manager_class,
                    (Package('tree'), Package('nano')),
                    r"rpm -q --whatprovides tree nano \|\| yum install -y  tree nano && rpm -q --whatprovides tree nano",  # noqa: E501
                    'Complete!',
                )

        elif package_manager_class is tmt.package_managers.dnf.Dnf:
            if 'ubi/8' in container.url:
                yield (
                    container,
                    package_manager_class,
                    (Package('dconf'), Package('libpng')),
                    r"rpm -q --whatprovides dconf libpng \|\| dnf install -y  dconf libpng",
                    'Complete!',
                )

            elif 'centos' in container.url:
                yield (
                    container,
                    package_manager_class,
                    (Package('tree'), Package('diffutils')),
                    r"rpm -q --whatprovides tree diffutils \|\| dnf install -y  tree diffutils",
                    'Complete!',
                )

            else:
                yield (
                    container,
                    package_manager_class,
                    (Package('tree'), Package('nano')),
                    r"rpm -q --whatprovides tree nano \|\| dnf install -y  tree nano",
                    'Complete!',
                )

        elif package_manager_class is tmt.package_managers.dnf.Dnf5:
            yield (
                container,
                package_manager_class,
                (Package('tree'), Package('nano')),
                r"rpm -q --whatprovides tree nano \|\| dnf5 install -y  tree nano",
                None,
            )

        elif package_manager_class is tmt.package_managers.apt.Apt:
            yield (
                container,
                package_manager_class,
                (Package('tree'), Package('nano')),
                r"set -x\s+export DEBIAN_FRONTEND=noninteractive\s+installable_packages=\"tree nano\"\s+dpkg-query --show \$installable_packages \\\s+\|\| apt install -y  \$installable_packages\s+exit \$\?",  # noqa: E501
                'Setting up tree',
            )

        elif package_manager_class is tmt.package_managers.rpm_ostree.RpmOstree:
            yield (
                container,
                package_manager_class,
                (Package('tree'), Package('nano')),
                r"rpm -q --whatprovides tree nano \|\| rpm-ostree install --apply-live --idempotent --allow-inactive --assumeyes  tree nano",  # noqa: E501
                'Installing: tree',
            )

        elif package_manager_class is tmt.package_managers.apk.Apk:
            yield (
                container,
                package_manager_class,
                (Package('tree'), Package('diffutils')),
                r"apk info -e tree diffutils \|\| apk add tree diffutils",
                'Installing tree',
            )

        else:
            pytest.fail(f"Unhandled package manager class '{package_manager_class}'.")


@pytest.mark.containers
@pytest.mark.parametrize(
    (
        'container_per_test',
        'package_manager_class',
        'packages',
        'expected_command',
        'expected_output',
    ),
    list(_parametrize_test_install_multiple()),
    indirect=["container_per_test"],
    ids=CONTAINER_MATRIX_IDS,
)
def test_install_multiple(
    container_per_test: ContainerData,
    guest_per_test: GuestContainer,
    package_manager_class: PackageManagerClass,
    packages: tuple[Package, Package],
    expected_command: str,
    expected_output: Optional[str],
    root_logger: tmt.log.Logger,
    caplog: _pytest.logging.LogCaptureFixture,
) -> None:
    package_manager = create_package_manager(
        container_per_test, guest_per_test, package_manager_class, root_logger
    )

    output = package_manager.install(*packages)

    assert_log(
        caplog,
        message=MATCH(rf"(?sm)Run command: podman exec .+? /bin/bash -c '{expected_command}'"),
    )

    assert_output(expected_output, output.stdout, output.stderr)


def _parametrize_test_install_downloaded() -> Iterator[
    tuple[
        Container,
        PackageManagerClass,
        tuple[Package, Package],
        tuple[str, str],
        str,
        Optional[str],
    ]
]:
    for container, package_manager_class in CONTAINER_BASE_MATRIX:
        if package_manager_class is tmt.package_managers.dnf.Yum:
            if 'centos/7' in container.url:
                yield pytest.param(
                    container,
                    package_manager_class,
                    (Package('tree'), Package('nano')),
                    ('tree*.x86_64.rpm', 'nano*.x86_64.rpm'),
                    r"yum install -y --skip-broken /tmp/tree.rpm /tmp/nano.rpm \|\| /bin/true",
                    'Complete!',
                    marks=pytest.mark.skip(reason="CentOS 7 does not support 'download' command"),
                )

            elif 'ubi/8' in container.url:
                yield (
                    container,
                    package_manager_class,
                    (Package('dconf'), Package('libpng')),
                    ('dconf*.x86_64.rpm', 'libpng*.x86_64.rpm'),
                    r"yum install -y --skip-broken /tmp/dconf.rpm /tmp/libpng.rpm \|\| /bin/true",
                    'Complete!',
                )

            elif 'centos' in container.url:
                yield (
                    container,
                    package_manager_class,
                    (Package('tree'), Package('diffutils')),
                    ('tree*.x86_64.rpm', 'diffutils*.x86_64.rpm'),
                    r"yum install -y --skip-broken /tmp/tree.rpm /tmp/diffutils.rpm \|\| /bin/true",  # noqa: E501
                    'Complete!',
                )

            else:
                yield (
                    container,
                    package_manager_class,
                    (Package('tree'), Package('nano')),
                    ('tree*.x86_64.rpm', 'nano*.x86_64.rpm'),
                    r"yum install -y --skip-broken /tmp/tree.rpm /tmp/nano.rpm \|\| /bin/true",
                    'Complete!',
                )

        elif package_manager_class is tmt.package_managers.dnf.Dnf:
            if 'ubi/8' in container.url:
                yield (
                    container,
                    package_manager_class,
                    (Package('dconf'), Package('libpng')),
                    ('dconf*.x86_64.rpm', 'libpng*.x86_64.rpm'),
                    r"dnf install -y  /tmp/dconf.rpm /tmp/libpng.rpm",
                    'Complete!',
                )

            elif 'centos/stream9' in container.url:
                yield (
                    container,
                    package_manager_class,
                    (Package('tree'), Package('diffutils')),
                    ('tree*.x86_64.rpm', 'diffutils*.x86_64.rpm'),
                    r"dnf install -y  /tmp/tree.rpm /tmp/diffutils.rpm",
                    'Complete!',
                )

            else:
                yield (
                    container,
                    package_manager_class,
                    (Package('tree'), Package('nano')),
                    ('tree*.x86_64.rpm', 'nano*.x86_64.rpm'),
                    r"dnf install -y  /tmp/tree.rpm /tmp/nano.rpm",
                    'Complete!',
                )

        elif package_manager_class is tmt.package_managers.dnf.Dnf5:
            yield (
                container,
                package_manager_class,
                (Package('tree'), Package('nano')),
                ('tree*.x86_64.rpm', 'nano*.x86_64.rpm'),
                r"dnf5 install -y  /tmp/tree.rpm /tmp/nano.rpm",
                None,
            )

        elif package_manager_class is tmt.package_managers.rpm_ostree.RpmOstree:
            yield (
                container,
                package_manager_class,
                (Package('tree'), Package('cowsay')),
                ('tree*.x86_64.rpm', 'cowsay*.noarch.rpm'),
                r"rpm-ostree install --apply-live --idempotent --allow-inactive --assumeyes  /tmp/tree.rpm /tmp/cowsay.rpm",  # noqa: E501
                'Installing: tree',
            )

        elif package_manager_class is tmt.package_managers.apt.Apt:
            yield pytest.param(
                container,
                package_manager_class,
                (Package('tree'), Package('nano')),
                ('tree*.x86_64.rpm', 'nano*.x86_64.rpm'),
                r".*?dpkg-query --show \$installable_packages \\\n^\|\| apt install -y  \$installable_packages.*",  # noqa: E501
                'Setting up tree',
                marks=pytest.mark.skip(reason="not supported yet"),
            )

        elif package_manager_class is tmt.package_managers.apk.Apk:
            yield pytest.param(
                container,
                package_manager_class,
                (Package('tree'), Package('nano')),
                ('tree*.x86_64.rpm', 'nano*.x86_64.rpm'),
                r"apk info -e tree nano \|\| apk add tree nano",
                'Installing tree',
                marks=pytest.mark.skip(reason="not supported yet"),
            )

        else:
            pytest.fail(f"Unhandled package manager class '{package_manager_class}'.")


@pytest.mark.containers
@pytest.mark.parametrize(
    (
        'container_per_test',
        'package_manager_class',
        'packages',
        'artifacts',
        'expected_command',
        'expected_output',
    ),
    list(_parametrize_test_install_downloaded()),
    indirect=["container_per_test"],
    ids=CONTAINER_MATRIX_IDS,
)
def test_install_downloaded(
    container_per_test: ContainerData,
    guest_per_test: GuestContainer,
    package_manager_class: PackageManagerClass,
    packages: tuple[Package, Package],
    artifacts: tuple[str, str],
    expected_command: str,
    expected_output: Optional[str],
    root_logger: tmt.log.Logger,
    caplog: _pytest.logging.LogCaptureFixture,
) -> None:
    package_manager = create_package_manager(
        container_per_test, guest_per_test, package_manager_class, root_logger
    )

    installables = tuple(PackagePath(f'/tmp/{package}.rpm') for package in packages)

    # TODO: move to a fixture
    guest_per_test.execute(
        ShellScript(
            f"""
        (yum download --destdir /tmp {packages[0]} {packages[1]} \
        || (dnf install -y 'dnf-command(download)' && dnf download --destdir /tmp {packages[0]} {packages[1]}) \
        || (dnf5 install -y 'dnf-command(download)' && dnf5 download --destdir /tmp {packages[0]} {packages[1]})) \
        && mv /tmp/{artifacts[0]} /tmp/{packages[0]}.rpm && mv /tmp/{artifacts[1]} /tmp/{packages[1]}.rpm
        """  # noqa: E501
        )
    )

    # TODO: yum and downloaded packages results in post-install `rpm -q`
    # check to make sure packages were indeed installed - but that
    # breaks because that test uses original package paths, not package
    # names, and fails. Disable this check for now, but it's a sloppy
    # "solution".
    if package_manager_class is tmt.package_managers.dnf.Yum:
        output = package_manager.install(
            *installables, options=Options(check_first=False, skip_missing=True)
        )

    else:
        output = package_manager.install(*installables, options=Options(check_first=False))

    assert_log(
        caplog,
        message=MATCH(rf"(?sm)Run command: podman exec .+? /bin/bash -c '{expected_command}'"),
    )

    assert_output(expected_output, output.stdout, output.stderr)


def _parametrize_test_install_debuginfo() -> Iterator[
    tuple[Container, PackageManagerClass, tuple[Package, Package], str, Optional[str]]
]:
    for container, package_manager_class in CONTAINER_BASE_MATRIX:
        # Skip testing debuginfo install on coreos images
        if 'coreos' in container.url:
            yield pytest.param(
                container,
                package_manager_class,
                (Package('tree'), Package('dos2unix')),
                "",
                None,
                marks=pytest.mark.skip(reason="debuginfo install not supported yet on coreos"),
            )

        elif package_manager_class is tmt.package_managers.dnf.Dnf5:
            yield (
                container,
                package_manager_class,
                (Package('dos2unix'), Package('tree')),
                r"rpm -q --whatprovides /usr/bin/debuginfo-install \|\| dnf5 install -y  /usr/bin/debuginfo-install && debuginfo-install -y  dos2unix tree && rpm -q dos2unix-debuginfo tree-debuginfo",  # noqa: E501
                None,
            )

        elif package_manager_class is tmt.package_managers.dnf.Dnf:
            if 'centos' in container.url:
                yield pytest.param(
                    container,
                    package_manager_class,
                    (Package('dos2unix'), Package('tree')),
                    r"debuginfo-install -y  dos2unix tree && rpm -q dos2unix-debuginfo tree-debuginfo",  # noqa: E501
                    None,
                    marks=pytest.mark.skip(
                        reason='centos comes without debuginfo repos, we do not enable them yet'
                    ),
                )

            elif 'ubi/8' in container.url:
                yield (
                    container,
                    package_manager_class,
                    (Package('dconf'), Package('libpng')),
                    r"rpm -q --whatprovides /usr/bin/debuginfo-install \|\| dnf install -y  /usr/bin/debuginfo-install && debuginfo-install -y  dconf libpng && rpm -q dconf-debuginfo libpng-debuginfo",  # noqa: E501
                    None,
                )

            else:
                yield (
                    container,
                    package_manager_class,
                    (Package('dos2unix'), Package('tree')),
                    r"rpm -q --whatprovides /usr/bin/debuginfo-install \|\| dnf install -y  /usr/bin/debuginfo-install && debuginfo-install -y  dos2unix tree && rpm -q dos2unix-debuginfo tree-debuginfo",  # noqa: E501
                    None,
                )

        elif package_manager_class is tmt.package_managers.dnf.Yum:
            if 'centos' in container.url:
                yield pytest.param(
                    container,
                    package_manager_class,
                    (Package('dos2unix'), Package('tree')),
                    r"debuginfo-install -y  dos2unix tree && rpm -q dos2unix-debuginfo tree-debuginfo",  # noqa: E501
                    None,
                    marks=pytest.mark.skip(
                        reason='centos comes without debuginfo repos, we do not enable them yet'
                    ),
                )

            elif 'ubi/8' in container.url:
                yield (
                    container,
                    package_manager_class,
                    (Package('dconf'), Package('libpng')),
                    r"rpm -q --whatprovides /usr/bin/debuginfo-install \|\| yum install -y  /usr/bin/debuginfo-install && rpm -q --whatprovides /usr/bin/debuginfo-install && debuginfo-install -y  dconf libpng && rpm -q dconf-debuginfo libpng-debuginfo",  # noqa: E501
                    None,
                )

            else:
                yield (
                    container,
                    package_manager_class,
                    (Package('dos2unix'), Package('tree')),
                    r"rpm -q --whatprovides /usr/bin/debuginfo-install \|\| yum install -y  /usr/bin/debuginfo-install && rpm -q --whatprovides /usr/bin/debuginfo-install && debuginfo-install -y  dos2unix tree && rpm -q dos2unix-debuginfo tree-debuginfo",  # noqa: E501
                    None,
                )

        elif (
            package_manager_class is tmt.package_managers.apt.Apt
            or package_manager_class is tmt.package_managers.rpm_ostree.RpmOstree
            or package_manager_class is tmt.package_managers.apk.Apk
        ):
            yield pytest.param(
                container,
                package_manager_class,
                (Package('tree'), Package('dos2unix')),
                "",
                None,
                marks=pytest.mark.skip(reason="not supported yet"),
            )

        else:
            pytest.fail(f"Unhandled package manager class '{package_manager_class}'.")


@pytest.mark.containers
@pytest.mark.parametrize(
    (
        'container_per_test',
        'package_manager_class',
        'installables',
        'expected_command',
        'expected_output',
    ),
    list(_parametrize_test_install_debuginfo()),
    indirect=["container_per_test"],
    ids=CONTAINER_MATRIX_IDS,
)
def test_install_debuginfo(
    container_per_test: ContainerData,
    guest_per_test: GuestContainer,
    package_manager_class: PackageManagerClass,
    installables: tuple[Package, Package],
    expected_command: str,
    expected_output: Optional[str],
    root_logger: tmt.log.Logger,
    caplog: _pytest.logging.LogCaptureFixture,
) -> None:
    package_manager = create_package_manager(
        container_per_test, guest_per_test, package_manager_class, root_logger
    )

    output = package_manager.install_debuginfo(*installables)

    assert_log(
        caplog, message=MATCH(rf"Run command: podman exec .+? /bin/bash -c '{expected_command}'")
    )

    assert_output(expected_output, output.stdout, output.stderr)


def _parametrize_test_install_debuginfo_nonexistent() -> Iterator[
    tuple[Container, PackageManagerClass, tuple[Package, Package], str, Optional[str]]
]:
    for container, package_manager_class in CONTAINER_BASE_MATRIX:
        if package_manager_class is tmt.package_managers.dnf.Dnf5:
            yield (
                container,
                package_manager_class,
                (Package('dos2unix'), Package('tree-but-spelled-wrong')),
                r"rpm -q --whatprovides /usr/bin/debuginfo-install || dnf5 install -y  /usr/bin/debuginfo-install && debuginfo-install -y  dos2unix tree-but-spelled-wrong && rpm -q dos2unix-debuginfo tree-but-spelled-wrong-debuginfo",  # noqa: E501
                None,
            )

        elif package_manager_class is tmt.package_managers.dnf.Dnf:
            yield (
                container,
                package_manager_class,
                (Package('dos2unix'), Package('tree-but-spelled-wrong')),
                r"rpm -q --whatprovides /usr/bin/debuginfo-install || dnf install -y  /usr/bin/debuginfo-install && debuginfo-install -y  dos2unix tree-but-spelled-wrong && rpm -q dos2unix-debuginfo tree-but-spelled-wrong-debuginfo",  # noqa: E501
                None,
            )

        elif package_manager_class is tmt.package_managers.dnf.Yum:
            yield (
                container,
                package_manager_class,
                (Package('dos2unix'), Package('tree-but-spelled-wrong')),
                r"rpm -q --whatprovides /usr/bin/debuginfo-install || yum install -y  /usr/bin/debuginfo-install && debuginfo-install -y  dos2unix tree-but-spelled-wrong && rpm -q dos2unix-debuginfo tree-but-spelled-wrong-debuginfo",  # noqa: E501
                None,
            )

        elif (
            package_manager_class is tmt.package_managers.apt.Apt
            or package_manager_class is tmt.package_managers.rpm_ostree.RpmOstree
            or package_manager_class is tmt.package_managers.apk.Apk
        ):
            yield pytest.param(
                container,
                package_manager_class,
                (Package('tree-but-spelled-wrong'), Package('dos2unix')),
                "",
                None,
                marks=pytest.mark.skip(reason="not supported yet"),
            )

        else:
            pytest.fail(f"Unhandled package manager class '{package_manager_class}'.")


@pytest.mark.containers
@pytest.mark.parametrize(
    (
        'container_per_test',
        'package_manager_class',
        'installables',
        'expected_command',
        'expected_output',
    ),
    list(_parametrize_test_install_debuginfo_nonexistent()),
    indirect=["container_per_test"],
    ids=CONTAINER_MATRIX_IDS,
)
def test_install_debuginfo_nonexistent(
    container_per_test: ContainerData,
    guest_per_test: GuestContainer,
    package_manager_class: PackageManagerClass,
    installables: tuple[Package, Package],
    expected_command: str,
    expected_output: Optional[str],
    root_logger: tmt.log.Logger,
    caplog: _pytest.logging.LogCaptureFixture,
) -> None:
    package_manager = create_package_manager(
        container_per_test, guest_per_test, package_manager_class, root_logger
    )

    with pytest.raises(tmt.utils.RunError) as excinfo:
        package_manager.install_debuginfo(*installables)

    assert_log(
        caplog, message=MATCH(rf"Run command: podman exec .+? /bin/bash -c '{expected_command}'")
    )

    assert excinfo.type is tmt.utils.RunError
    assert excinfo.value.returncode != 0

    assert_output(expected_output, excinfo.value.stdout, excinfo.value.stderr)


def _parametrize_test_install_debuginfo_nonexistent_skip() -> Iterator[
    tuple[Container, PackageManagerClass, tuple[Package, Package], str, Optional[str]]
]:
    for container, package_manager_class in CONTAINER_BASE_MATRIX:
        # Skip testing debuginfo install on coreos images
        if 'coreos' in container.url:
            yield pytest.param(
                container,
                package_manager_class,
                (Package('tree'), Package('dos2unix')),
                "",
                None,
                marks=pytest.mark.skip(reason="debuginfo install not supported yet on coreos"),
            )

        elif package_manager_class is tmt.package_managers.dnf.Dnf5:
            yield (
                container,
                package_manager_class,
                (Package('dos2unix'), Package('tree-but-spelled-wrong')),
                r"rpm -q --whatprovides /usr/bin/debuginfo-install \|\| dnf5 install -y  /usr/bin/debuginfo-install && debuginfo-install -y --skip-broken dos2unix tree-but-spelled-wrong",  # noqa: E501
                None,
            )

        elif package_manager_class is tmt.package_managers.dnf.Dnf:
            if 'centos' in container.url:
                yield pytest.param(
                    container,
                    package_manager_class,
                    (Package('dos2unix'), Package('tree-but-spelled-wrong')),
                    r"debuginfo-install -y --skip-broken dos2unix tree-but-spelled-wrong",
                    None,
                    marks=pytest.mark.skip(
                        reason='centos comes without debuginfo repos, we do not enable them yet'
                    ),
                )

            else:
                yield (
                    container,
                    package_manager_class,
                    (Package('dos2unix'), Package('tree-but-spelled-wrong')),
                    r"rpm -q --whatprovides /usr/bin/debuginfo-install \|\| dnf install -y  /usr/bin/debuginfo-install && debuginfo-install -y --skip-broken dos2unix tree-but-spelled-wrong",  # noqa: E501
                    None,
                )

        elif package_manager_class is tmt.package_managers.dnf.Yum:
            if 'centos' in container.url:
                yield pytest.param(
                    container,
                    package_manager_class,
                    (Package('dos2unix'), Package('tree-but-spelled-wrong')),
                    r"debuginfo-install -y --skip-broken dos2unix tree-but-spelled-wrong",
                    None,
                    marks=pytest.mark.skip(
                        reason='centos comes without debuginfo repos, we do not enable them yet'
                    ),
                )

            else:
                yield (
                    container,
                    package_manager_class,
                    (Package('dos2unix'), Package('tree-but-spelled-wrong')),
                    r"rpm -q --whatprovides /usr/bin/debuginfo-install || yum install -y  /usr/bin/debuginfo-install && rpm -q --whatprovides /usr/bin/debuginfo-install && debuginfo-install -y --skip-broken dos2unix tree-but-spelled-wrong",  # noqa: E501
                    None,
                )

        elif (
            package_manager_class is tmt.package_managers.apt.Apt
            or package_manager_class is tmt.package_managers.rpm_ostree.RpmOstree
            or package_manager_class is tmt.package_managers.apk.Apk
        ):
            yield pytest.param(
                container,
                package_manager_class,
                (Package('tree-but-spelled-wrong'), Package('dos2unix')),
                "",
                None,
                marks=pytest.mark.skip(reason="not supported yet"),
            )

        else:
            pytest.fail(f"Unhandled package manager class '{package_manager_class}'.")


@pytest.mark.containers
@pytest.mark.parametrize(
    (
        'container_per_test',
        'package_manager_class',
        'installables',
        'expected_command',
        'expected_output',
    ),
    list(_parametrize_test_install_debuginfo_nonexistent_skip()),
    indirect=["container_per_test"],
    ids=CONTAINER_MATRIX_IDS,
)
def test_install_debuginfo_nonexistent_skip(
    container_per_test: ContainerData,
    guest_per_test: GuestContainer,
    package_manager_class: PackageManagerClass,
    installables: tuple[Package, Package],
    expected_command: str,
    expected_output: Optional[str],
    root_logger: tmt.log.Logger,
    caplog: _pytest.logging.LogCaptureFixture,
) -> None:
    package_manager = create_package_manager(
        container_per_test, guest_per_test, package_manager_class, root_logger
    )

    output = package_manager.install_debuginfo(*installables, options=Options(skip_missing=True))

    assert_log(
        caplog, message=MATCH(rf"Run command: podman exec .+? /bin/bash -c '{expected_command}'")
    )

    assert_output(expected_output, output.stdout, output.stderr)



import asyncio
import dataclasses
import datetime
import importlib.metadata
import logging
import os
import re
from collections.abc import Mapping
from contextlib import suppress
from functools import wraps
from typing import Any, Callable, Optional, TypedDict, Union, cast

import packaging.version

import tmt
import tmt.config
import tmt.hardware
import tmt.log
import tmt.options
import tmt.steps
import tmt.steps.provision
import tmt.utils
import tmt.utils.signals
import tmt.utils.wait
from tmt.config.models.hardware import MrackTranslation
from tmt.container import container, field, simple_field
from tmt.utils import (
    Command,
    Path,
    ProvisionError,
    ShellScript,
    UpdatableMessage,
)
from tmt.utils.templates import render_template
from tmt.utils.wait import Deadline, Waiting

MRACK_VERSION: Optional[str] = None

mrack: Any
providers: Any
ProvisioningError: Any
NotAuthenticatedError: Any
BEAKER: Any
BeakerProvider: Any
BeakerTransformer: Any
TmtBeakerTransformer: Any

_MRACK_IMPORTED: bool = False

DEFAULT_USER = 'root'
DEFAULT_ARCH = 'x86_64'
DEFAULT_IMAGE = 'fedora'
DEFAULT_PROVISION_TIMEOUT = 3600  # 1 hour timeout at least
DEFAULT_PROVISION_TICK = 60  # poll job each minute

#: How often Beaker session should be refreshed to pick up up-to-date
#: Kerberos ticket.
DEFAULT_API_SESSION_REFRESH = 3600


def mrack_constructs_ks_pre() -> bool:
    """
    Kickstart construction has been improved in 1.21.0
    """

    assert MRACK_VERSION is not None

    return packaging.version.Version(MRACK_VERSION) >= packaging.version.Version('1.21.0')


def _get_constraint_translations(logger: tmt.log.Logger) -> list[MrackTranslation]:
    """
    Load the list of hardware requirement translations from configuration.

    :returns: translations loaded from configuration, or an empty list if
        the configuration is missing.
    """
    config = tmt.config.Config(logger)
    return (
        config.hardware.beaker.translations if config.hardware and config.hardware.beaker else []
    )


# Type annotation for "data" package describing a guest instance. Passed
# between load() and save() calls


class GuestInspectType(TypedDict):
    status: str
    system: str
    address: Optional[str]


# Mapping of HW requirement operators to their Beaker representation.
OPERATOR_SIGN_TO_OPERATOR = {
    tmt.hardware.Operator.EQ: '==',
    tmt.hardware.Operator.NEQ: '!=',
    tmt.hardware.Operator.GT: '>',
    tmt.hardware.Operator.GTE: '>=',
    tmt.hardware.Operator.LT: '<',
    tmt.hardware.Operator.LTE: '<=',
}


def operator_to_beaker_op(operator: tmt.hardware.Operator, value: str) -> tuple[str, str, bool]:
    """
    Convert constraint operator to Beaker "op".

    :param operator: operator to convert.
    :param value: value operator works with. It shall be a string representation
        of the the constraint value, as converted for the Beaker job XML.
    :returns: tuple of three items: Beaker operator, fit for ``op`` attribute
        of XML filters, a value to go with it instead of the input one, and
        a boolean signalizing whether the filter, constructed by the caller,
        should be negated.
    """

    if operator in OPERATOR_SIGN_TO_OPERATOR:
        return OPERATOR_SIGN_TO_OPERATOR[operator], value, False

    # MATCH has special handling - convert the pattern to a wildcard form -
    # and that may be weird :/
    if operator == tmt.hardware.Operator.MATCH:
        return 'like', value.replace('.*', '%').replace('.+', '%'), False

    if operator == tmt.hardware.Operator.NOTMATCH:
        return 'like', value.replace('.*', '%').replace('.+', '%'), True

    raise ProvisionError(f"Hardware requirement operator '{operator}' is not supported.")


# Transcription of our HW constraints into Mrack's own representation. It's based
# on dictionaries, and it's slightly weird. There is no distinction between elements
# that do not have attributes, like <and/>, and elements that must have them, like
# <memory/> and other binary operations. Also, there is no distinction between
# element attribute and child element, both are specified as dictionary key, just
# the former would be a string, the latter another, nested, dictionary.
#
# This makes it harder for us to enforce correct structure of the transcribed tree.
# Therefore adding a thin layer of containers that describe what Mrack is willing
# to accept, but with strict type annotations; the layer is aware of how to convert
# its components into dictionaries.
@container
class MrackBaseHWElement:
    """
    Base for Mrack hardware requirement elements
    """

    # Only a name is defined, as it's the only property shared across all element
    # types.
    name: str

    def to_mrack(self) -> dict[str, Any]:
        """
        Convert the element to Mrack-compatible dictionary tree
        """

        raise NotImplementedError


@container
class MrackHWElement(MrackBaseHWElement):
    """
    An element with name and attributes.

    This type of element is not allowed to have any child elements.
    """

    attributes: dict[str, str] = simple_field(default_factory=dict)

    def to_mrack(self) -> dict[str, Any]:
        return {self.name: self.attributes}


@container(init=False)
class MrackHWBinOp(MrackHWElement):
    """
    An element describing a binary operation, a "check"
    """

    def __init__(self, name: str, operator: str, value: str) -> None:
        super().__init__(name)

        self.attributes = {'_op': operator, '_value': value}


@container(init=False)
class MrackHWKeyValue(MrackHWElement):
    """
    A key-value element
    """

    def __init__(self, name: str, operator: str, value: str) -> None:
        super().__init__('key_value')

        self.attributes = {'_key': name, '_op': operator, '_value': value}


BeakerizedConstraint = Union[MrackBaseHWElement, dict[str, Any]]


@container
class MrackHWGroup(MrackBaseHWElement):
    """
    An element with child elements.

    This type of element is not allowed to have any attributes.
    """

    children: list[BeakerizedConstraint] = simple_field(default_factory=list)

    def to_mrack(self) -> dict[str, Any]:
        # Another unexpected behavior of mrack dictionary tree: if there is just
        # a single child, it is "packed" into its parent as a key/dict item.
        if (
            len(self.children) == 1
            and self.name not in ('and', 'or')
            and not isinstance(self.children[0], dict)
        ):
            return {self.name: self.children[0].to_mrack()}

        return {
            self.name: [
                child.to_mrack() if isinstance(child, MrackBaseHWElement) else child
                for child in self.children
            ]
        }


@container
class MrackHWAndGroup(MrackHWGroup):
    """
    Represents ``<and/>`` element
    """

    name: str = 'and'


@container
class MrackHWOrGroup(MrackHWGroup):
    """
    Represents ``<or/>`` element
    """

    name: str = 'or'


@container
class MrackHWNotGroup(MrackHWGroup):
    """
    Represents ``<not/>`` element
    """

    name: str = 'not'


def _transform_unsupported(constraint: tmt.hardware.Constraint) -> dict[str, Any]:
    # We have to return something, return an {}- no harm done, composable with other elements.

    return {}


def _translate_constraint_by_config(
    constraint: tmt.hardware.Constraint,
    logger: tmt.log.Logger,
) -> dict[str, Any]:
    """
    Translate hardware constraints to Mrack-compatible dictionary tree with config.
    """

    config = _get_constraint_translations(logger)

    if not config:
        return _transform_unsupported(constraint)

    suitable_translations = [
        translation
        for translation in config
        if translation.requirement == constraint.printable_name
    ]

    if not suitable_translations:
        return _transform_unsupported(constraint)

    beaker_operator, actual_value, negate = operator_to_beaker_op(
        constraint.operator, str(constraint.value)
    )

    return tmt.utils.yaml_to_dict(
        render_template(
            suitable_translations[0].template,
            CONSTRAINT=constraint,
            BEAKER_OPERATOR=beaker_operator,
            BEAKER_VALUE=actual_value,
            BEAKER_NEGATE=negate,
        )
    )


def _translate_constraint_by_transformer(
    constraint: tmt.hardware.Constraint,
    logger: tmt.log.Logger,
) -> BeakerizedConstraint:
    """
    Translate hardware constraints to Mrack-compatible dictionary tree with transformer.
    """

    name, _, child_name = constraint.expand_name()

    if child_name:
        transformer = _CONSTRAINT_TRANSFORMERS.get(f'{name}.{child_name}')

    else:
        transformer = _CONSTRAINT_TRANSFORMERS.get(name)
    if not transformer:
        return _transform_unsupported(constraint)
    return transformer(constraint, logger)


def _transform_beaker_pool(
    constraint: tmt.hardware.TextConstraint, logger: tmt.log.Logger
) -> MrackBaseHWElement:
    beaker_operator, actual_value, negate = operator_to_beaker_op(
        constraint.operator, constraint.value
    )

    return MrackHWBinOp('pool', beaker_operator, actual_value)


def _transform_cpu_family(
    constraint: tmt.hardware.IntegerConstraint, logger: tmt.log.Logger
) -> MrackBaseHWElement:
    beaker_operator, actual_value, _ = operator_to_beaker_op(
        constraint.operator, str(constraint.value)
    )

    return MrackHWGroup('cpu', children=[MrackHWBinOp('family', beaker_operator, actual_value)])


def _transform_cpu_flag(
    constraint: tmt.hardware.TextConstraint, logger: tmt.log.Logger
) -> MrackBaseHWElement:
    beaker_operator = (
        OPERATOR_SIGN_TO_OPERATOR[tmt.hardware.Operator.EQ]
        if constraint.operator is tmt.hardware.Operator.CONTAINS
        else OPERATOR_SIGN_TO_OPERATOR[tmt.hardware.Operator.NEQ]
    )
    actual_value = str(constraint.value)

    return MrackHWGroup('cpu', children=[MrackHWBinOp('flag', beaker_operator, actual_value)])


def _transform_cpu_model(
    constraint: tmt.hardware.IntegerConstraint, logger: tmt.log.Logger
) -> MrackBaseHWElement:
    beaker_operator, actual_value, _ = operator_to_beaker_op(
        constraint.operator, str(constraint.value)
    )

    return MrackHWGroup('cpu', children=[MrackHWBinOp('model', beaker_operator, actual_value)])


def _transform_cpu_processors(
    constraint: tmt.hardware.IntegerConstraint, logger: tmt.log.Logger
) -> MrackBaseHWElement:
    beaker_operator, actual_value, _ = operator_to_beaker_op(
        constraint.operator, str(constraint.value)
    )

    return MrackHWGroup(
        'cpu', children=[MrackHWBinOp('processors', beaker_operator, actual_value)]
    )


def _transform_cpu_cores(
    constraint: tmt.hardware.IntegerConstraint, logger: tmt.log.Logger
) -> MrackBaseHWElement:
    beaker_operator, actual_value, _ = operator_to_beaker_op(
        constraint.operator, str(constraint.value)
    )

    return MrackHWGroup('cpu', children=[MrackHWBinOp('cores', beaker_operator, actual_value)])


def _transform_cpu_model_name(
    constraint: tmt.hardware.TextConstraint, logger: tmt.log.Logger
) -> MrackBaseHWElement:
    beaker_operator, actual_value, negate = operator_to_beaker_op(
        constraint.operator, constraint.value
    )

    if negate:
        return MrackHWNotGroup(
            children=[
                MrackHWGroup(
                    'cpu', children=[MrackHWBinOp('model_name', beaker_operator, actual_value)]
                )
            ]
        )

    return MrackHWGroup(
        'cpu', children=[MrackHWBinOp('model_name', beaker_operator, actual_value)]
    )


def _transform_cpu_frequency(
    constraint: tmt.hardware.NumberConstraint, logger: tmt.log.Logger
) -> MrackBaseHWElement:
    beaker_operator, actual_value, _ = operator_to_beaker_op(
        constraint.operator, str(float(constraint.value.to('MHz').magnitude))
    )

    return MrackHWGroup('cpu', children=[MrackHWBinOp('speed', beaker_operator, actual_value)])


def _transform_cpu_stepping(
    constraint: tmt.hardware.IntegerConstraint, logger: tmt.log.Logger
) -> MrackBaseHWElement:
    beaker_operator, actual_value, _ = operator_to_beaker_op(
        constraint.operator, str(constraint.value)
    )

    return MrackHWGroup('cpu', children=[MrackHWBinOp('stepping', beaker_operator, actual_value)])


def _transform_cpu_vendor_name(
    constraint: tmt.hardware.TextConstraint, logger: tmt.log.Logger
) -> MrackBaseHWElement:
    beaker_operator, actual_value, negate = operator_to_beaker_op(
        constraint.operator, str(constraint.value)
    )

    if negate:
        return MrackHWNotGroup(
            children=[
                MrackHWGroup(
                    'cpu', children=[MrackHWBinOp('vendor', beaker_operator, actual_value)]
                )
            ]
        )

    return MrackHWGroup('cpu', children=[MrackHWBinOp('vendor', beaker_operator, actual_value)])


def _transform_cpu_hyper_threading(
    constraint: tmt.hardware.FlagConstraint, logger: tmt.log.Logger
) -> MrackBaseHWElement:
    beaker_operator, actual_value, _ = operator_to_beaker_op(
        constraint.operator, str(constraint.value)
    )

    return MrackHWGroup('cpu', children=[MrackHWBinOp('hyper', beaker_operator, actual_value)])


def _transform_disk_driver(
    constraint: tmt.hardware.TextConstraint, logger: tmt.log.Logger
) -> MrackBaseHWElement:
    beaker_operator, actual_value, negate = operator_to_beaker_op(
        constraint.operator, constraint.value
    )

    if negate:
        return MrackHWNotGroup(
            children=[MrackHWKeyValue('BOOTDISK', beaker_operator, actual_value)]
        )

    return MrackHWKeyValue('BOOTDISK', beaker_operator, actual_value)


def _transform_disk_size(
    constraint: tmt.hardware.SizeConstraint, logger: tmt.log.Logger
) -> MrackBaseHWElement:
    beaker_operator, actual_value, _ = operator_to_beaker_op(
        constraint.operator, str(int(constraint.value.to('B').magnitude))
    )

    return MrackHWGroup('disk', children=[MrackHWBinOp('size', beaker_operator, actual_value)])


def _transform_disk_model_name(
    constraint: tmt.hardware.TextConstraint, logger: tmt.log.Logger
) -> MrackBaseHWElement:
    beaker_operator, actual_value, negate = operator_to_beaker_op(
        constraint.operator, constraint.value
    )

    if negate:
        return MrackHWNotGroup(
            children=[
                MrackHWGroup(
                    'disk', children=[MrackHWBinOp('model', beaker_operator, actual_value)]
                )
            ]
        )

    return MrackHWGroup('disk', children=[MrackHWBinOp('model', beaker_operator, actual_value)])


def _transform_disk_physical_sector_size(
    constraint: tmt.hardware.SizeConstraint, logger: tmt.log.Logger
) -> MrackBaseHWElement:
    beaker_operator, actual_value, _ = operator_to_beaker_op(
        constraint.operator, str(constraint.value)
    )

    return MrackHWGroup(
        'disk', children=[MrackHWBinOp('phys_sector_size', beaker_operator, actual_value)]
    )


def _transform_disk_logical_sector_size(
    constraint: tmt.hardware.SizeConstraint, logger: tmt.log.Logger
) -> MrackBaseHWElement:
    beaker_operator, actual_value, _ = operator_to_beaker_op(
        constraint.operator, str(constraint.value)
    )

    return MrackHWGroup(
        'disk', children=[MrackHWBinOp('sector_size', beaker_operator, actual_value)]
    )


def _transform_hostname(
    constraint: tmt.hardware.TextConstraint, logger: tmt.log.Logger
) -> MrackBaseHWElement:
    beaker_operator, actual_value, negate = operator_to_beaker_op(
        constraint.operator, constraint.value
    )

    if negate:
        return MrackHWNotGroup(children=[MrackHWBinOp('hostname', beaker_operator, actual_value)])

    return MrackHWBinOp('hostname', beaker_operator, actual_value)


def _transform_memory(
    constraint: tmt.hardware.SizeConstraint, logger: tmt.log.Logger
) -> MrackBaseHWElement:
    beaker_operator, actual_value, _ = operator_to_beaker_op(
        constraint.operator, str(int(constraint.value.to('MiB').magnitude))
    )

    return MrackHWGroup('system', children=[MrackHWBinOp('memory', beaker_operator, actual_value)])


def _transform_tpm_version(
    constraint: tmt.hardware.TextConstraint, logger: tmt.log.Logger
) -> MrackBaseHWElement:
    beaker_operator, actual_value, negate = operator_to_beaker_op(
        constraint.operator, constraint.value
    )

    return MrackHWKeyValue('TPM', beaker_operator, actual_value)


def _transform_virtualization_is_virtualized(
    constraint: tmt.hardware.FlagConstraint, logger: tmt.log.Logger
) -> BeakerizedConstraint:
    beaker_operator, actual_value, _ = operator_to_beaker_op(
        constraint.operator, str(constraint.value)
    )

    test = (constraint.operator, constraint.value)

    if test in [(tmt.hardware.Operator.EQ, True), (tmt.hardware.Operator.NEQ, False)]:
        return MrackHWGroup('system', children=[MrackHWBinOp('hypervisor', '!=', '')])

    if test in [(tmt.hardware.Operator.EQ, False), (tmt.hardware.Operator.NEQ, True)]:
        return MrackHWGroup('system', children=[MrackHWBinOp('hypervisor', '==', '')])

    return _transform_unsupported(constraint)


def _transform_virtualization_hypervisor(
    constraint: tmt.hardware.TextConstraint, logger: tmt.log.Logger
) -> MrackBaseHWElement:
    beaker_operator, actual_value, negate = operator_to_beaker_op(
        constraint.operator, str(constraint.value)
    )

    if negate:
        return MrackHWNotGroup(
            children=[
                MrackHWGroup(
                    'system', children=[MrackHWBinOp('hypervisor', beaker_operator, actual_value)]
                )
            ]
        )

    return MrackHWGroup(
        'system', children=[MrackHWBinOp('hypervisor', beaker_operator, actual_value)]
    )


def _transform_zcrypt_adapter(
    constraint: tmt.hardware.TextConstraint, logger: tmt.log.Logger
) -> MrackBaseHWElement:
    beaker_operator, actual_value, negate = operator_to_beaker_op(
        constraint.operator, constraint.value
    )

    if negate:
        return MrackHWNotGroup(
            children=[
                MrackHWGroup(
                    'system',
                    children=[MrackHWKeyValue('ZCRYPT_MODEL', beaker_operator, actual_value)],
                )
            ]
        )

    return MrackHWGroup(
        'system', children=[MrackHWKeyValue('ZCRYPT_MODEL', beaker_operator, actual_value)]
    )


def _transform_zcrypt_mode(
    constraint: tmt.hardware.TextConstraint, logger: tmt.log.Logger
) -> MrackBaseHWElement:
    beaker_operator, actual_value, negate = operator_to_beaker_op(
        constraint.operator, constraint.value
    )

    if negate:
        return MrackHWNotGroup(
            children=[
                MrackHWGroup(
                    'system',
                    children=[MrackHWKeyValue('ZCRYPT_MODE', beaker_operator, actual_value)],
                )
            ]
        )

    return MrackHWGroup(
        'system', children=[MrackHWKeyValue('ZCRYPT_MODE', beaker_operator, actual_value)]
    )


def _transform_iommu_is_supported(
    constraint: tmt.hardware.FlagConstraint, logger: tmt.log.Logger
) -> BeakerizedConstraint:
    test = (constraint.operator, constraint.value)

    if test in [(tmt.hardware.Operator.EQ, True), (tmt.hardware.Operator.NEQ, False)]:
        return MrackHWKeyValue('VIRT_IOMMU', '==', '1')

    if test in [(tmt.hardware.Operator.EQ, False), (tmt.hardware.Operator.NEQ, True)]:
        return MrackHWKeyValue('VIRT_IOMMU', '==', '0')

    return _transform_unsupported(constraint)


def _transform_location_lab_controller(
    constraint: tmt.hardware.TextConstraint, logger: tmt.log.Logger
) -> MrackBaseHWElement:
    if constraint.operator not in [tmt.hardware.Operator.EQ, tmt.hardware.Operator.NEQ]:
        raise ProvisionError(
            f"Cannot apply hardware requirement '{constraint}', operator not supported."
        )
    beaker_operator, actual_value, negate = operator_to_beaker_op(
        constraint.operator, constraint.value
    )

    if negate:
        return MrackHWNotGroup(
            children=[MrackHWBinOp('labcontroller', beaker_operator, actual_value)]
        )

    return MrackHWBinOp('labcontroller', beaker_operator, actual_value)


def _transform_system_numa_nodes(
    constraint: tmt.hardware.IntegerConstraint, logger: tmt.log.Logger
) -> MrackBaseHWElement:
    beaker_operator, actual_value, _ = operator_to_beaker_op(
        constraint.operator, str(constraint.value)
    )

    return MrackHWGroup(
        'system', children=[MrackHWBinOp('numanodes', beaker_operator, actual_value)]
    )


def _transform_system_model_name(
    constraint: tmt.hardware.TextConstraint, logger: tmt.log.Logger
) -> MrackBaseHWElement:
    beaker_operator, actual_value, negate = operator_to_beaker_op(
        constraint.operator, str(constraint.value)
    )

    if negate:
        return MrackHWNotGroup(
            children=[
                MrackHWGroup(
                    'system', children=[MrackHWBinOp('model', beaker_operator, actual_value)]
                )
            ]
        )

    return MrackHWGroup('system', children=[MrackHWBinOp('model', beaker_operator, actual_value)])


def _transform_system_vendor_name(
    constraint: tmt.hardware.TextConstraint, logger: tmt.log.Logger
) -> MrackBaseHWElement:
    beaker_operator, actual_value, negate = operator_to_beaker_op(
        constraint.operator, str(constraint.value)
    )

    if negate:
        return MrackHWNotGroup(
            children=[
                MrackHWGroup(
                    'system', children=[MrackHWBinOp('vendor', beaker_operator, actual_value)]
                )
            ]
        )

    return MrackHWGroup('system', children=[MrackHWBinOp('vendor', beaker_operator, actual_value)])


ConstraintTransformer = Callable[[tmt.hardware.Constraint, tmt.log.Logger], MrackBaseHWElement]

_CONSTRAINT_TRANSFORMERS: Mapping[str, ConstraintTransformer] = {
    'beaker.pool': _transform_beaker_pool,  # type: ignore[dict-item]
    'cpu.cores': _transform_cpu_cores,  # type: ignore[dict-item]
    'cpu.family': _transform_cpu_family,  # type: ignore[dict-item]
    'cpu.flag': _transform_cpu_flag,  # type: ignore[dict-item]
    'cpu.frequency': _transform_cpu_frequency,  # type: ignore[dict-item]
    'cpu.hyper_threading': _transform_cpu_hyper_threading,  # type: ignore[dict-item]
    'cpu.model': _transform_cpu_model,  # type: ignore[dict-item]
    'cpu.model_name': _transform_cpu_model_name,  # type: ignore[dict-item]
    'cpu.processors': _transform_cpu_processors,  # type: ignore[dict-item]
    'cpu.stepping': _transform_cpu_stepping,  # type: ignore[dict-item]
    'cpu.vendor_name': _transform_cpu_vendor_name,  # type: ignore[dict-item]
    'disk.driver': _transform_disk_driver,  # type: ignore[dict-item]
    'disk.model_name': _transform_disk_model_name,  # type: ignore[dict-item]
    'disk.size': _transform_disk_size,  # type: ignore[dict-item]
    'disk.physical_sector_size': _transform_disk_physical_sector_size,  # type: ignore[dict-item]
    'disk.logical_sector_size': _transform_disk_logical_sector_size,  # type: ignore[dict-item]
    'hostname': _transform_hostname,  # type: ignore[dict-item]
    'location.lab_controller': _transform_location_lab_controller,  # type: ignore[dict-item]
    'memory': _transform_memory,  # type: ignore[dict-item]
    'tpm.version': _transform_tpm_version,  # type: ignore[dict-item]
    'virtualization.is_virtualized': _transform_virtualization_is_virtualized,  # type: ignore[dict-item]
    'virtualization.hypervisor': _transform_virtualization_hypervisor,  # type: ignore[dict-item]
    'zcrypt.adapter': _transform_zcrypt_adapter,  # type: ignore[dict-item]
    'zcrypt.mode': _transform_zcrypt_mode,  # type: ignore[dict-item]
    'system.numa_nodes': _transform_system_numa_nodes,  # type: ignore[dict-item]
    'system.model_name': _transform_system_model_name,  # type: ignore[dict-item]
    'system.vendor_name': _transform_system_vendor_name,  # type: ignore[dict-item]
    'iommu.is_supported': _transform_iommu_is_supported,  # type: ignore[dict-item]
}


def constraint_to_beaker_filter(
    constraint: tmt.hardware.BaseConstraint, logger: tmt.log.Logger
) -> BeakerizedConstraint:
    """
    Convert a hardware constraint into a Mrack-compatible filter
    """

    if isinstance(constraint, tmt.hardware.And):
        return MrackHWAndGroup(
            children=[
                constraint_to_beaker_filter(child_constraint, logger)
                for child_constraint in constraint.constraints
            ]
        )

    if isinstance(constraint, tmt.hardware.Or):
        return MrackHWOrGroup(
            children=[
                constraint_to_beaker_filter(child_constraint, logger)
                for child_constraint in constraint.constraints
            ]
        )

    assert isinstance(constraint, tmt.hardware.Constraint)

    transformed = _translate_constraint_by_config(
        constraint, logger
    ) or _translate_constraint_by_transformer(constraint, logger)

    if not transformed:
        # Unsupported constraint has been already logged via report_support().
        # Make sure user is aware it would have no effect.
        logger.warning(f"Hardware requirement '{constraint.printable_name}' will have no effect.")
    return transformed


def import_and_load_mrack_deps(workdir: Any, name: str, logger: tmt.log.Logger) -> None:
    """
    Import mrack module only when needed
    """

    global _MRACK_IMPORTED

    if _MRACK_IMPORTED:
        return

    global MRACK_VERSION
    global mrack
    global providers
    global ProvisioningError
    global NotAuthenticatedError
    global BEAKER
    global BeakerProvider
    global BeakerTransformer
    global TmtBeakerTransformer

    try:
        import mrack
        from mrack.errors import NotAuthenticatedError, ProvisioningError
        from mrack.providers import providers
        from mrack.providers.beaker import PROVISIONER_KEY as BEAKER
        from mrack.providers.beaker import BeakerProvider
        from mrack.transformers.beaker import BeakerTransformer

        MRACK_VERSION = importlib.metadata.version('mrack')

        # hack: remove mrack stdout and move the logfile to /tmp
        mrack.logger.removeHandler(mrack.console_handler)
        mrack.logger.removeHandler(mrack.file_handler)

        with suppress(OSError):
            os.remove("mrack.log")

        logging.FileHandler(str(f"{workdir}/{name}-mrack.log"))

        providers.register(BEAKER, BeakerProvider)

    except ImportError:
        raise ProvisionError("Install 'tmt+provision-beaker' to provision using this method.")

    # ignore the misc because mrack sources are not typed and result into
    # error: Class cannot subclass "BeakerTransformer" (has type "Any")
    # as mypy does not have type information for the BeakerTransformer class
    class TmtBeakerTransformer(BeakerTransformer):  # type: ignore[misc]
        def _translate_tmt_hw(self, hw: tmt.hardware.Hardware) -> dict[str, Any]:
            """
            Return hw requirements from given hw dictionary
            """

            assert hw.constraint
            # Beaker, unlike instance-type-based infrastructures like AWS, does
            # have the actual filtering, and can express `or` and `and`
            # groups. And our `constraint_to_beaker_filter()` does that,
            # even for groups nested deeper in the tree.
            transformed = constraint_to_beaker_filter(hw.constraint, logger)
            if isinstance(transformed, MrackBaseHWElement):
                transformed = transformed.to_mrack()

            logger.debug('Transformed hardware', tmt.utils.dict_to_yaml(transformed))

            # Mrack does not handle well situation when the filter
            # consists of just a single filtering element, e.g. just
            # `hostname`. In that case, the element is converted into
            # XML element incorrectly. Therefore wrapping our filter
            # with `<and/>` group, even if it has just a single child,
            # it works around the problem.
            # See https://github.com/teemtee/tmt/issues/3442
            return {'hostRequires': MrackHWAndGroup(children=[transformed]).to_mrack()}

        def create_host_requirement(self, host: CreateJobParameters) -> dict[str, Any]:
            """
            Create single input for Beaker provisioner
            """

            req: dict[str, Any] = super().create_host_requirement(host.to_mrack())

            if host.hardware and host.hardware.constraint:
                req.update(self._translate_tmt_hw(host.hardware))

            if host.beaker_job_owner:
                req['job_owner'] = host.beaker_job_owner

            if host.kickstart:
                if 'kernel-options' in host.kickstart:
                    req['kernel_options'] = host.kickstart['kernel-options']

                if 'kernel-options-post' in host.kickstart:
                    req['kernel_options_post'] = host.kickstart['kernel-options-post']

                if not mrack_constructs_ks_pre():
                    ks_components: list[str] = []

                    for ks_section in ('pre-install', 'script', 'post-install'):
                        if ks_section in host.kickstart:
                            ks_components.append(host.kickstart[ks_section])

                    if ks_components:
                        req['ks_append'] = ['\n'.join(ks_components)]

            # Whiteboard must be added *after* request preparation, to overwrite the default one.
            req['whiteboard'] = host.whiteboard

            logger.debug('mrack request', req, level=4)

            logger.info('whiteboard', host.whiteboard, 'green')

            return req

    _MRACK_IMPORTED = True


def async_run(func: Any) -> Any:
    """
    Decorate click actions to run as async
    """

    @wraps(func)
    def update_wrapper(*args: Any, **kwargs: Any) -> Any:
        return asyncio.run(func(*args, **kwargs))

    return update_wrapper


@container
class BeakerGuestData(tmt.steps.provision.GuestSshData):
    # Override parent class with our defaults
    user: str = field(
        default=DEFAULT_USER,
        option=('-u', '--user'),
        metavar='USERNAME',
        help='Username to use for all guest operations.',
    )

    # Guest request properties
    whiteboard: Optional[str] = field(
        default=None,
        option=('-w', '--whiteboard'),
        metavar='WHITEBOARD',
        help='Text description of the beaker job which is displayed in the list of jobs.',
    )
    arch: str = field(
        default=DEFAULT_ARCH,
        option='--arch',
        metavar='ARCH',
        help='Architecture to provision.',
    )
    image: Optional[str] = field(
        default=DEFAULT_IMAGE,
        option=('-i', '--image'),
        metavar='COMPOSE',
        help='Image (distro or "compose" in Beaker terminology) to provision.',
    )

    # Provided in Beaker job
    job_id: Optional[str] = field(
        default=None,
        internal=True,
    )

    # Timeouts and deadlines
    provision_timeout: int = field(
        default=DEFAULT_PROVISION_TIMEOUT,
        option='--provision-timeout',
        metavar='SECONDS',
        help=f"""
             How long to wait for provisioning to complete,
             {DEFAULT_PROVISION_TIMEOUT} seconds by default.
             """,
        normalize=tmt.utils.normalize_int,
    )
    provision_tick: int = field(
        default=DEFAULT_PROVISION_TICK,
        option='--provision-tick',
        metavar='SECONDS',
        help=f"""
             How often check Beaker for provisioning status,
             {DEFAULT_PROVISION_TICK} seconds by default.
             """,
        normalize=tmt.utils.normalize_int,
    )
    api_session_refresh_tick: int = field(
        default=DEFAULT_API_SESSION_REFRESH,
        option='--api-session-refresh-tick',
        metavar='SECONDS',
        help=f"""
             How often should Beaker session be refreshed to pick up-to-date Kerberos ticket,
             {DEFAULT_API_SESSION_REFRESH} seconds by default.
             """,
        normalize=tmt.utils.normalize_int,
    )
    kickstart: dict[str, str] = field(
        default_factory=dict,
        option='--kickstart',
        metavar='KEY=VALUE',
        help='Optional Beaker kickstart to use when provisioning the guest.',
        multiple=True,
        normalize=tmt.utils.normalize_string_dict,
    )

    beaker_job_owner: Optional[str] = field(
        default=None,
        option='--beaker-job-owner',
        metavar='USERNAME',
        help="""
             If set, Beaker jobs will be submitted on behalf of ``USERNAME``.
             Submitting user must be a submission delegate for the ``USERNAME``.
             """,
    )

    public_key: list[str] = field(
        default_factory=list,
        option='--public-key',
        metavar='PUBKEY',
        help="""
             Public keys to add among authorized SSH keys.
             """,
        multiple=True,
        normalize=tmt.utils.normalize_string_list,
    )

    beaker_job_group: Optional[str] = field(
        default=None,
        option='--beaker-job-group',
        metavar='GROUPNAME',
        help="""
             If set, Beaker jobs will be submitted on behalf of ``GROUPNAME``.
             """,
    )


@container
class ProvisionBeakerData(BeakerGuestData, tmt.steps.provision.ProvisionStepData):
    pass


GUEST_STATE_COLOR_DEFAULT = 'green'

GUEST_STATE_COLORS = {
    "New": "blue",
    "Scheduled": "blue",
    "Queued": "cyan",
    "Processed": "cyan",
    "Waiting": "magenta",
    "Installing": "magenta",
    "Running": "magenta",
    "Cancelled": "yellow",
    "Aborted": "yellow",
    "Reserved": "green",
    "Completed": "green",
}


@container
class CreateJobParameters:
    """
    Collect all parameters for a future Beaker job
    """

    tmt_name: str
    name: str
    os: str
    arch: str
    hardware: Optional[tmt.hardware.Hardware]
    kickstart: dict[str, str]
    whiteboard: Optional[str]
    beaker_job_owner: Optional[str]
    public_key: list[str]
    group: Optional[str]

    def to_mrack(self) -> dict[str, Any]:
        data = dataclasses.asdict(self)

        data['beaker'] = {}

        if self.kickstart:
            kickstart = self.kickstart.copy()

            if 'metadata' in kickstart:
                data['beaker']['ks_meta'] = kickstart.pop('metadata')

            # Mrack does not handle metadata-only kickstart nicely, ends
            # up with just an empty string. Don't tempt it, don't let it
            # see kickstart if it was just metadata.
            if kickstart:
                data['beaker']['ks_append'] = kickstart
        if self.public_key:
            data['beaker']['pubkeys'] = self.public_key

        return data


class BeakerAPI:
    # req is a requirement passed to Beaker mrack provisioner
    mrack_requirement: dict[str, Any] = {}
    dsp_name: str = "Beaker"

    # wrapping around the __init__ with async wrapper does mangle the method
    # and mypy complains as it no longer returns None but the coroutine
    @async_run
    async def __init__(self, guest: 'GuestBeaker') -> None:  # type: ignore[misc]
        """
        Initialize the API class with defaults and load the config
        """

        self._guest = guest

        # use global context class
        global_context = mrack.context.global_context

        mrack_config_locations = [
            Path(__file__).parent / "mrack/mrack.conf",
            Path("/etc/tmt/mrack.conf"),
            Path("~/.mrack/mrack.conf").expanduser(),
            Path.cwd() / "mrack.conf",
        ]

        mrack_config: Optional[Path] = None

        for potential_location in mrack_config_locations:
            if potential_location.exists():
                mrack_config = potential_location

        if not mrack_config:
            raise ProvisionError("Configuration file 'mrack.conf' not found.")

        try:
            global_context.init(str(mrack_config))
        except mrack.errors.ConfigError as mrack_conf_err:
            raise ProvisionError(mrack_conf_err)

        self._mrack_transformer = TmtBeakerTransformer()
        try:
            await self._mrack_transformer.init(global_context.PROV_CONFIG, {})
        except NotAuthenticatedError as kinit_err:
            raise ProvisionError(kinit_err) from kinit_err
        except AttributeError as hub_err:
            raise ProvisionError(
                f"Can not use current kerberos ticket to authenticate: {hub_err}"
            ) from hub_err
        except FileNotFoundError as missing_conf_err:
            raise ProvisionError(
                f"Configuration file missing: {missing_conf_err.filename}"
            ) from missing_conf_err

        self._mrack_provider = self._mrack_transformer._provider
        self._mrack_provider.poll_sleep = DEFAULT_PROVISION_TICK

        if guest.job_id:
            self._bkr_job_id = guest.job_id

    @async_run
    async def create(self, data: CreateJobParameters) -> Any:
        """
        Create - or request creation of - a resource using mrack up.

        :param data: describes the provisioning request.
        """

        mrack_requirement = self._mrack_transformer.create_host_requirement(data)
        log_msg_start = f"{self.dsp_name} [{self.mrack_requirement.get('name')}]"
        self._bkr_job_id, self._req = await self._mrack_provider.create_server(mrack_requirement)
        return self._mrack_provider._get_recipe_info(self._bkr_job_id, log_msg_start)

    @async_run
    async def inspect(
        self,
    ) -> Any:
        """
        Inspect a resource (kinda wait till provisioned)
        """

        log_msg_start = f"{self.dsp_name} [{self.mrack_requirement.get('name')}]"
        return self._mrack_provider._get_recipe_info(self._bkr_job_id, log_msg_start)

    @async_run
    async def delete(  # destroy
        self,
    ) -> Any:
        """
        Delete - or request removal of - a resource
        """

        return await self._mrack_provider.delete_host(self._bkr_job_id, None)


class GuestBeaker(tmt.steps.provision.GuestSsh):
    """
    Beaker guest instance
    """

    _data_class = BeakerGuestData

    # Guest request properties
    whiteboard: Optional[str]
    arch: str
    image: str = "fedora-latest"
    hardware: Optional[tmt.hardware.Hardware] = None
    kickstart: dict[str, str]

    beaker_job_owner: Optional[str] = None
    beaker_job_group: Optional[str] = None

    # Provided in Beaker response
    job_id: Optional[str]

    # Timeouts and deadlines
    provision_timeout: int
    provision_tick: int
    public_key: list[str]
    api_session_refresh_tick: int

    _api: Optional[BeakerAPI] = None
    _api_timestamp: Optional[datetime.datetime] = None

    @property
    def api(self) -> BeakerAPI:
        """
        Create BeakerAPI leveraging mrack
        """

        def _construct_api() -> tuple[BeakerAPI, datetime.datetime]:
            assert self.parent is not None

            import_and_load_mrack_deps(self.parent.workdir, self.parent.name, self._logger)

            return BeakerAPI(self), datetime.datetime.now(datetime.timezone.utc)

        if self._api is None:
            self._api, self._api_timestamp = _construct_api()

        else:
            assert self._api_timestamp is not None

            delta = datetime.datetime.now(datetime.timezone.utc) - self._api_timestamp

            if delta.total_seconds() >= self.api_session_refresh_tick:
                self.debug(f'Refresh Beaker API client as it is too old, {delta}.')

                self._api, self._api_timestamp = _construct_api()

        return self._api

    @property
    def is_ready(self) -> bool:
        """
        Check if provisioning of machine is done
        """

        if self.job_id is None:
            return False

        assert mrack is not None

        try:
            response = self.api.inspect()

            if response["status"] == "Aborted":
                return False

            current = cast(GuestInspectType, response)
            state = current["status"]
            if state in {"Error, Aborted", "Cancelled"}:
                return False

            if state == 'Reserved':
                return True
            return False

        except mrack.errors.MrackError:
            return False

    def _create(self, tmt_name: str) -> None:
        """
        Create beaker job xml request and submit it to Beaker hub
        """

        data = CreateJobParameters(
            tmt_name=tmt_name,
            hardware=self.hardware,
            kickstart=self.kickstart,
            arch=self.arch,
            os=self.image,
            name=f'{self.image}-{self.arch}',
            whiteboard=self.whiteboard or tmt_name,
            beaker_job_owner=self.beaker_job_owner,
            public_key=self.public_key,
            group=self.beaker_job_group,
        )

        if self.is_dry_run:
            mrack_requirement = self.api._mrack_transformer.create_host_requirement(data)
            job = self.api._mrack_provider._req_to_bkr_job(mrack_requirement)

            # Mrack indents XML with tabs, tmt indents with spaces, let's make sure we
            # don't mix these two in tmt output & modify mrack's output to use spaces as well.
            self.print(
                re.sub(
                    r'^\t+',
                    lambda match: '  ' * len(match.group()),
                    job.toxml(prettyxml=True),
                    flags=re.MULTILINE,
                ),
                shift=-2,
            )
            return

        with tmt.utils.signals.PreventSignals(self._logger):
            try:
                response = self.api.create(data)

            except ProvisioningError as exc:
                import xmlrpc.client

                cause = exc.__cause__

                if isinstance(cause, xmlrpc.client.Fault):
                    if 'is not a valid user name' in cause.faultString:
                        raise ProvisionError(
                            f"Failed to create Beaker job, job owner '{self.beaker_job_owner}' "
                            "was refused as unknown."
                        ) from exc

                    if 'is not a valid submission delegate' in cause.faultString:
                        raise ProvisionError(
                            f"Failed to create Beaker job, job owner '{self.beaker_job_owner}' "
                            "is not a valid submission delegate."
                        ) from exc

                    if 'is not a valid group' in cause.faultString:
                        raise ProvisionError(
                            f"Failed to create Beaker job, job group '{self.beaker_job_group}' "
                            "was refused as unknown."
                        ) from exc

                    if 'is not a member of group' in cause.faultString:
                        raise ProvisionError(
                            "Failed to create Beaker job, submitting user is not "
                            "a member of group '{self.beaker_job_group}'"
                        ) from exc

                raise ProvisionError('Failed to create Beaker job') from exc

            if not response:
                raise ProvisionError(f"Failed to create, response: '{response}'.")

            self.info('guest', 'has been requested', 'green')
            self.job_id = f'J:{response["id"]}'

        self.info('job id', self.job_id, 'green')

        with UpdatableMessage("status", indent_level=self._level()) as progress_message:

            def get_new_state() -> GuestInspectType:
                response = self.api.inspect()

                if response["status"] == "Aborted":
                    raise ProvisionError(
                        f"Failed to create, unhandled API response '{response['status']}'."
                    )

                current = cast(GuestInspectType, response)
                state = current["status"]
                state_color = GUEST_STATE_COLORS.get(state, GUEST_STATE_COLOR_DEFAULT)

                progress_message.update(state, color=state_color)

                if state in {"Error, Aborted", "Cancelled"}:
                    raise ProvisionError('Failed to create, provisioning failed.')

                if state == 'Reserved':
                    for key in response.get('logs', []):
                        self.guest_logs.append(
                            GuestLogBeaker(key.replace('.log', ''), self, response["logs"][key])
                        )
                    # console.log contains dmesg, and accessible even when the system is dead.
                    if response.get('logs', []).get('console.log'):
                        self.guest_logs.append(
                            GuestLogBeaker('dmesg', self, response.get('logs').get('console.log'))
                        )
                    else:
                        self.warn('No console.log available.')
                    return current

                raise tmt.utils.wait.WaitingIncompleteError

            try:
                guest_info = Waiting(
                    Deadline.from_seconds(self.provision_timeout), tick=self.provision_tick
                ).wait(get_new_state, self._logger)

            except tmt.utils.wait.WaitingTimedOutError:
                response = self.api.delete()
                raise ProvisionError(
                    f'Failed to provision in the given amount '
                    f'of time (--provision-timeout={self.provision_timeout}).'
                )

        self.primary_address = self.topology_address = guest_info['system']

    def start(self) -> None:
        """
        Start the guest

        Get a new guest instance running. This should include preparing
        any configuration necessary to get it started. Called after
        load() is completed so all guest data should be available.
        """

        if self.job_id is None or self.primary_address is None:
            self._create(self._tmt_name())

        self.verbose('primary address', self.primary_address, 'green')
        self.verbose('topology address', self.topology_address, 'green')

    def remove(self) -> None:
        """
        Remove the guest
        """

        if self.job_id is None:
            return

        self.api.delete()

    def reboot(
        self,
        hard: bool = False,
        command: Optional[Union[Command, ShellScript]] = None,
        waiting: Optional[Waiting] = None,
    ) -> bool:
        """
        Reboot the guest, and wait for the guest to recover.

        .. note::

           Custom reboot command can be used only in combination with a
           soft reboot. If both ``hard`` and ``command`` are set, a hard
           reboot will be requested, and ``command`` will be ignored.

        :param hard: if set, force the reboot. This may result in a loss
            of data. The default of ``False`` will attempt a graceful
            reboot.

            Plugin will use ``bkr system-power`` command to trigger the
            hard reboot. Unlike ``command``, this command would be
            executed on the runner, **not** on the guest.
        :param command: a command to run on the guest to trigger the
            reboot. If ``hard`` is also set, ``command`` is ignored.
        :param timeout: amount of time in which the guest must become available
            again.
        :param tick: how many seconds to wait between two consecutive attempts
            of contacting the guest.
        :param tick_increase: a multiplier applied to ``tick`` after every
            attempt.
        :returns: ``True`` if the reboot succeeded, ``False`` otherwise.
        """

        waiting = waiting or tmt.steps.provision.default_reboot_waiting()

        if hard:
            self.debug("Hard reboot using the reboot command 'bkr system-power --action reboot'.")

            reboot_script = ShellScript(f'bkr system-power --action reboot {self.primary_address}')

            return self.perform_reboot(
                lambda: self._run_guest_command(reboot_script.to_shell_command()),
                waiting,
                fetch_boot_time=False,
            )

        return super().reboot(
            hard=False,
            command=command,
            waiting=waiting,
        )


@tmt.steps.provides_method('beaker')
class ProvisionBeaker(tmt.steps.provision.ProvisionPlugin[ProvisionBeakerData]):
    """
    Provision guest on Beaker system using mrack.

    Reserve a machine from the Beaker pool using the ``mrack``
    plugin. ``mrack`` is a multicloud provisioning library
    supporting multiple cloud services including Beaker.

    The following two files are used for configuration:

    ``/etc/tmt/mrack.conf`` for basic configuration

    ``/etc/tmt/provisioning-config.yaml`` configuration per supported provider

    Beaker installs distribution specified by the ``image``
    key. If the image can not be translated using the
    ``provisioning-config.yaml`` file mrack passes the image
    value to Beaker hub and tries to request distribution
    based on the image value. This way we can bypass default
    translations and use desired distribution specified like
    the one in the example below.

    Minimal configuration could look like this:

    .. code-block:: yaml

        provision:
            how: beaker
            image: fedora

    To trigger a hard reboot of a guest, ``bkr system-power --action reboot``
    command is executed.

    .. warning::

        ``bkr system-power`` command is executed on the runner, not
        on the guest.

    .. code-block:: yaml

        # Specify the distro directly
        provision:
            how: beaker
            image: Fedora-37%

    .. code-block:: yaml

        # Set custom whiteboard description (added in 1.30)
        provision:
            how: beaker
            whiteboard: Just a smoke test for now
    """

    _data_class = ProvisionBeakerData
    _guest_class = GuestBeaker

    # _thread_safe = True

    # Guest instance
    _guest = None

    # data argument should be a "Optional[GuestData]" type but we would like to use
    # BeakerGuestData created here ignoring the override will make mypy calm
    def wake(self, data: Optional[BeakerGuestData] = None) -> None:  # type: ignore[override]
        """
        Wake up the plugin, process data, apply options
        """

        super().wake(data=data)

        if data:
            self._guest = GuestBeaker(
                data=data,
                name=self.name,
                parent=self.step,
                logger=self._logger,
            )

    def go(self, *, logger: Optional[tmt.log.Logger] = None) -> None:
        """
        Provision the guest
        """

        super().go(logger=logger)

        data = BeakerGuestData.from_plugin(self)

        data.show(verbose=self.verbosity_level, logger=self._logger)

        if data.hardware:
            data.hardware.report_support(
                names=list(_CONSTRAINT_TRANSFORMERS.keys()), logger=self._logger
            )

        self._guest = GuestBeaker(
            data=data,
            name=self.name,
            parent=self.step,
            logger=self._logger,
        )
        self._guest.start()
        self._guest.setup()


@container
class GuestLogBeaker(tmt.steps.provision.GuestLog):
    guest: GuestBeaker
    url: str

    def fetch(self, logger: tmt.log.Logger) -> Optional[str]:
        """
        Fetch and return content of a log.

        :returns: content of the log, or ``None`` if the log cannot be retrieved.
        """
        try:
            return tmt.utils.get_url_content(self.url)
        except Exception as error:
            logger.warning(f'Failed to fetch log: {error}')
            return None



import webbrowser
from typing import Optional

from jinja2 import select_autoescape

import tmt
import tmt.log
import tmt.steps
import tmt.steps.report
import tmt.utils
import tmt.utils.templates
from tmt.container import container, field
from tmt.utils import Path

HTML_TEMPLATE_PATH = tmt.utils.resource_files('steps/report/html/template.html.j2')
DEFAULT_NAME = 'index.html'

DEFAULT_FILENAME = 'index.html'


@container
class ReportHtmlData(tmt.steps.report.ReportStepData):
    file: Optional[Path] = field(
        default=None,
        option='--file',
        metavar='PATH',
        help='Path to the file to store HTML report to.',
        normalize=tmt.utils.normalize_path,
    )

    open: bool = field(
        default=False,
        option=('-o', '--open'),
        is_flag=True,
        help='Open results in your preferred web browser.',
    )

    absolute_paths: bool = field(
        default=False,
        option='--absolute-paths',
        is_flag=True,
        help='Make paths absolute rather than relative to working directory.',
    )

    display_guest: str = field(
        default='auto',
        option='--display-guest',
        metavar='auto|always|never',
        choices=['auto', 'always', 'never'],
        help="""
             When to display full guest name in report: when more than a single guest was involved
             (default), always, or never.
             """,
    )


@tmt.steps.provides_method('html')
class ReportHtml(tmt.steps.report.ReportPlugin[ReportHtmlData]):
    """
    Format test results into an HTML report.

    Create a local ``html`` file with test results arranged in
    a table. Optionally open the page in the default browser.

    Example config:

    .. code-block:: yaml

        # Enable html report from the command line
        tmt run --all report --how html
        tmt run --all report --how html --open
        tmt run -l report -h html -o

    .. code-block:: yaml

        # Use html as the default report for given plan
        report:
            how: html
            open: true
    """

    _data_class = ReportHtmlData

    @property
    def _preserved_workdir_members(self) -> set[str]:
        """
        A set of members of the step workdir that should not be removed.
        """

        return {DEFAULT_FILENAME}

    def go(self, *, logger: Optional[tmt.log.Logger] = None) -> None:
        """
        Process results
        """

        super().go(logger=logger)

        # Prepare the template
        environment = tmt.utils.templates.default_template_environment()

        # Explicitly enable the autoescape because it's disabled by default by TMT
        # (see /teemtee/tmt/issues/2873 for more info.
        environment.autoescape = select_autoescape(enabled_extensions=('html.j2'))

        if self.data.absolute_paths:

            def _linkable_path(path: str) -> str:
                return str(Path(path).absolute())

            environment.filters["linkable_path"] = _linkable_path
        else:
            # Links used in html should be relative to a workdir
            def _linkable_path(path: str) -> str:
                assert self.workdir is not None  # narrow type

                return str(Path(path).relative_to(self.workdir))

            environment.filters["linkable_path"] = _linkable_path

        if self.data.display_guest == 'always':
            display_guest = True

        elif self.data.display_guest == 'never':
            display_guest = False

        else:
            seen_guests = {result.guest.name for result in self.step.plan.execute.results()}

            display_guest = len(seen_guests) > 1

        # Write the report
        assert self.workdir is not None
        filepath = self.data.file or self.workdir / DEFAULT_FILENAME

        try:
            filepath.write_text(
                tmt.utils.templates.render_template_file(
                    HTML_TEMPLATE_PATH,
                    environment,
                    results=self.step.plan.execute.results(),
                    base_dir=self.step.plan.execute.workdir,
                    plan=self.step.plan,
                    display_guest=display_guest,
                ),
            )

        except Exception as error:
            raise tmt.utils.ReportError(f"Failed to write the output '{filepath}'.") from error

        # Nothing more to do in dry mode
        if self.is_dry_run:
            return

        # Show output file path
        self.info("output", filepath, color='yellow')
        if not self.data.open:
            return

        # Open target in webbrowser
        try:
            if webbrowser.open(f"file://{filepath}", new=0):
                self.info('open', 'Successfully opened in the web browser.', color='green')
                return
            self.fail("Failed to open the web browser.")
        except Exception as error:
            self.fail(f"Failed to open the web browser: {error}")

        raise tmt.utils.ReportError("Unable to open the web browser.")
